#!/usr/bin/env python
"""server.py: Process requests to RNNSearch"""
__author__ = "Frederic Bergeron"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "bergeron@pa.jst.jp"
__status__ = "Development"

import datetime
import json
import numpy as np
from chainer import cuda
import logging
import sys
import tempfile

from nmt_chainer.dataprocessing.processors import build_dataset_one_side_pp
import nmt_chainer.translation.eval
from nmt_chainer.translation.server_arg_parsing import make_config_server

import traceback

import time
import timeit
import socket
import threading
import SocketServer
import xml.etree.ElementTree as ET
import re
import subprocess
import bokeh.embed

logging.basicConfig()
log = logging.getLogger("rnns:server")
log.setLevel(logging.INFO)


class Translator:

    def __init__(self, config_server):
        self.config_server = config_server
        from nmt_chainer.translation.eval import create_encdec
        self.encdec, self.eos_idx, self.src_indexer, self.tgt_indexer, self.reverse_encdec = create_encdec(
            config_server)
        if 'gpu' in config_server.process and config_server.process.gpu is not None:
            self.encdec = self.encdec.to_gpu(config_server.process.gpu)

        self.encdec_list = [self.encdec]

    def translate(self, sentence, beam_width, beam_pruning_margin, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, nb_steps, nb_steps_ratio,
                  remove_unk, normalize_unicode_unk, attempt_to_relocate_unk_source, beam_score_length_normalization, beam_score_length_normalization_strength, post_score_length_normalization, post_score_length_normalization_strength,
                  post_score_coverage_penalty, post_score_coverage_penalty_strength,
                  groundhog, force_finish, prob_space_combination, attn_graph_width, attn_graph_height):
        from nmt_chainer.utilities import visualisation
        log.info("processing source string %s" % sentence)

        src_file = tempfile.NamedTemporaryFile()
        src_file.write(sentence.encode('utf-8'))
        src_file.seek(0)

        dest_file = tempfile.NamedTemporaryFile()
        rich_output_file = tempfile.NamedTemporaryFile()
        attn_graph_script_file = tempfile.NamedTemporaryFile()
        attn_graph_div_file = tempfile.NamedTemporaryFile()

        try:
            out = ''
            script = ''
            div = '<div/>'
            unk_mapping = []

            src_data, stats_src_pp = build_dataset_one_side_pp(src_file.name, self.src_indexer, max_nb_ex=self.config_server.process.max_nb_ex)

            from nmt_chainer.translation.eval import translate_to_file_with_beam_search
            translate_to_file_with_beam_search(dest_file.name, self.config_server.process.gpu, self.encdec, self.eos_idx, src_data, beam_width, beam_pruning_margin,
                                               beam_score_coverage_penalty=beam_score_coverage_penalty,
                                               beam_score_coverage_penalty_strength=beam_score_coverage_penalty_strength,
                                               nb_steps=nb_steps,
                                               nb_steps_ratio=nb_steps_ratio,
                                               beam_score_length_normalization=beam_score_length_normalization,
                                               beam_score_length_normalization_strength=beam_score_length_normalization_strength,
                                               post_score_length_normalization=post_score_length_normalization,
                                               post_score_length_normalization_strength=post_score_length_normalization_strength,
                                               post_score_coverage_penalty=post_score_coverage_penalty,
                                               post_score_coverage_penalty_strength=post_score_coverage_penalty_strength,
                                               groundhog=groundhog,
                                               tgt_unk_id=self.config_server.output.tgt_unk_id,
                                               tgt_indexer=self.tgt_indexer,
                                               force_finish=force_finish,
                                               prob_space_combination=prob_space_combination, reverse_encdec=self.reverse_encdec,
                                               generate_attention_html=(attn_graph_script_file.name, attn_graph_div_file.name),
                                               attn_graph_with_sum=False,
                                               attn_graph_attribs={'title': '', 'toolbar_location': 'below', 'plot_width': attn_graph_width, 'plot_height': attn_graph_height}, src_indexer=self.src_indexer,
                                               rich_output_filename=rich_output_file.name,
                                               use_unfinished_translation_if_none_found=False,
                                               replace_unk=True, src=sentence, dic=self.config_server.output.dic,
                                               remove_unk=remove_unk, normalize_unicode_unk=normalize_unicode_unk, attempt_to_relocate_unk_source=attempt_to_relocate_unk_source)

            dest_file.seek(0)
            out = dest_file.read()

            rich_output_file.seek(0)
            rich_output_data = json.loads(rich_output_file.read())
            unk_mapping = rich_output_data[0]['unk_mapping']

            attn_graph_script_file.seek(0)
            script = attn_graph_script_file.read()

            attn_graph_div_file.seek(0)
            div = attn_graph_div_file.read()

        finally:
            src_file.close()
            dest_file.close()
            rich_output_file.close()
            attn_graph_script_file.close()
            attn_graph_div_file.close()

        return out, script, div, unk_mapping


class RequestHandler(SocketServer.BaseRequestHandler):

    def handle(self):
        start_request = timeit.default_timer()
        log.info(timestamped_msg("Handling request..."))
        data = self.request.recv(4096)

        response = {}
        if (data):
            try:
                cur_thread = threading.current_thread()

                log.info("data={0}".format(data))
                root = ET.fromstring(data)
                article_id = root.get('id')
                try:
                    attn_graph_width = int(root.get('attn_graph_width', 0))
                except BaseException:
                    attn_graph_width = 0
                try:
                    attn_graph_height = int(root.get('attn_graph_height', 0))
                except BaseException:
                    attn_graph_height = 0
                beam_width = int(root.get('beam_width', 30))
                nb_steps = int(root.get('nb_steps', 50))
                beam_pruning_margin = None
                try:
                    beam_pruning_margin = float(root.get('beam_pruning_margin'))
                except BaseException:
                    pass
                beam_score_coverage_penalty = root.get(
                    'beam_score_coverage_penalty', 'none')
                beam_score_coverage_penalty_strength = None
                try:
                    beam_score_coverage_penalty_strength = float(root.get('beam_score_coverage_penalty_strength', 0.2))
                except BaseException:
                    pass
                nb_steps_ratio = None
                try:
                    nb_steps_ratio = float(root.get('nb_steps_ratio', 1.2))
                except BaseException:
                    pass
                groundhog = ('true' == root.get('groundhog', 'false'))
                force_finish = ('true' == root.get('force_finish', 'false'))
                beam_score_length_normalization = root.get(
                    'beam_score_length_normalization', 'none')
                beam_score_length_normalization_strength = None
                try:
                    beam_score_length_normalization_strength = float(root.get('beam_score_length_normalization_strength', 0.2))
                except BaseException:
                    pass
                post_score_length_normalization = root.get(
                    'post_score_length_normalization', 'simple')
                post_score_length_normalization_strength = None
                try:
                    post_score_length_normalization_strength = float(root.get('post_score_length_normalization_strength', 0.2))
                except BaseException:
                    pass
                post_score_coverage_penalty = root.get(
                    'post_score_coverage_penalty', 'none')
                post_score_coverage_penalty_strength = None
                try:
                    post_score_coverage_penalty_strength = float(root.get('post_score_coverage_penalty_strength', 0.2))
                except BaseException:
                    pass
                prob_space_combination = (
                    'true' == root.get(
                        'prob_space_combination', 'false'))
                remove_unk = ('true' == root.get('remove_unk', 'false'))
                normalize_unicode_unk = (
                    'true' == root.get(
                        'normalize_unicode_unk', 'true'))
                log.info('normalize_unicode_unk=' + str(normalize_unicode_unk))
                attempt_to_relocate_unk_source = ('true' == root.get(
                    'attempt_to_relocate_unk_source', 'false'))
                log.info("Article id: %s" % article_id)
                out = ""
                graph_data = []
                segmented_input = []
                segmented_output = []
                mapping = []
                sentences = root.findall('sentence')
                for idx, sentence in enumerate(sentences):
                    sentence_number = sentence.get('id')
                    text = sentence.findtext('i_sentence').strip()
                    log.info("text=@@@%s@@@" % text)

                    cmd = self.server.segmenter_command % text
                    log.info("cmd=%s" % cmd)
                    start_cmd = timeit.default_timer()

                    parser_output = subprocess.check_output(cmd, shell=True)

                    log.info(
                        "Segmenter request processed in {} s.".format(
                            timeit.default_timer() - start_cmd))
                    log.info("parser_output=%s" % parser_output)

                    words = []
                    if 'parse_server' == self.server.segmenter_format:
                        for line in parser_output.split("\n"):
                            if (line.startswith('#')):
                                continue
                            elif (not line.strip()):
                                break
                            else:
                                parts = line.split("\t")
                                word = parts[2]
                                words.append(word)
                    elif 'morph' == self.server.segmenter_format:
                        for pair in parser_output.split(' '):
                            if pair != '':
                                word, pos = pair.split('_')
                                words.append(word)
                    elif 'plain' == self.server.segmenter_format:
                        words = parser_output.split(' ')
                    else:
                        pass
                    splitted_sentence = ' '.join(words)
                    # log.info("splitted_sentence=" + splitted_sentence)

                    log.info(timestamped_msg("Translating sentence %d" % idx))
                    decoded_sentence = splitted_sentence.decode('utf-8')
                    translation, script, div, unk_mapping = self.server.translator.translate(decoded_sentence,
                                                                                             beam_width, beam_pruning_margin, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, nb_steps, nb_steps_ratio, remove_unk, normalize_unicode_unk, attempt_to_relocate_unk_source,
                                                                                             beam_score_length_normalization, beam_score_length_normalization_strength, post_score_length_normalization, post_score_length_normalization_strength, post_score_coverage_penalty, post_score_coverage_penalty_strength,
                                                                                             groundhog, force_finish, prob_space_combination, attn_graph_width, attn_graph_height)
                    out += translation
                    segmented_input.append(splitted_sentence)
                    segmented_output.append(translation)
                    mapping.append(unk_mapping)
                    graph_data.append(
                        (script.encode('utf-8'), div.encode('utf-8')))

                    # There should always be only one sentence for now. - FB
                    break

                response['article_id'] = article_id
                response['sentence_number'] = sentence_number
                response['out'] = out
                response['segmented_input'] = segmented_input
                response['segmented_output'] = segmented_output
                response['mapping'] = map(lambda x: ' '.join(x), mapping)
                graphes = []
                for gd in graph_data:
                    script, div = gd
                    graphes.append({'script': script, 'div': div})
                response['attn_graphes'] = graphes
            except BaseException:
                traceback.print_exc()
                error_lines = traceback.format_exc().splitlines()
                response['error'] = error_lines[-1]
                response['stacktrace'] = error_lines

        log.info(
            "Request processed in {0} s. by {1}".format(
                timeit.default_timer() -
                start_request,
                cur_thread.name))

        response = json.dumps(response)
        self.request.sendall(response)


class Server(SocketServer.ThreadingMixIn, SocketServer.TCPServer):

    daemon_threads = True
    allow_reuse_address = True

    def __init__(
            self,
            server_address,
            handler_class,
            segmenter_command,
            segmenter_format,
            translator):
        SocketServer.TCPServer.__init__(self, server_address, handler_class)
        self.segmenter_command = segmenter_command
        self.segmenter_format = segmenter_format
        self.translator = translator


def timestamped_msg(msg):
    timestamp = datetime.datetime.fromtimestamp(time.time()).strftime('%Y-%m-%d %H:%M:%S')
    return "{0}: {1}".format(timestamp, msg)


def do_start_server(config_server):
    translator = Translator(config_server)
    server_host, server_port = config_server.process.server.split(":")
    server = Server(
        (server_host,
         int(server_port)),
        RequestHandler,
        config_server.process.segmenter_command,
        config_server.process.segmenter_format,
        translator)
    ip, port = server.server_address
    log.info(
        timestamped_msg(
            "Start listening for requests on {0}:{1}...".format(
                socket.gethostname(),
                port)))

    try:
        server.serve_forever()
    except KeyboardInterrupt:
        server.shutdown()
        server.server_close()

    sys.exit(0)


if __name__ == '__main__':
    command_line()



#!/usr/bin/env python
"""models.py: Implementation of RNNSearch in Chainer"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import numpy as np
import chainer
from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L

import rnn_cells
from nmt_chainer.utilities.utils import compute_lexicon_matrix
import attention

import logging
logging.basicConfig()
log = logging.getLogger("rnns:models")
log.setLevel(logging.INFO)


class BNList(ChainList):
    def __init__(self, size, max_length):
        super(BNList, self).__init__()
        for _ in xrange(max_length):
            bn = L.BatchNormalization(size)
            bn.gamma.data.fill(0.1)
            self.add_link(bn)
        self.max_length = max_length

    def __call__(self, x, i, test=False):
        if i < self.max_length:
            return self[i](x, test=test)
        else:
            return self[self.max_length - 1](x, test=test)


class GradKeeper(chainer.Function):
    def __init__(self, grad):
        super(GradKeeper, self).__init__()
        self.grad = grad

    def forward(self, inputs):
        return (np.array([0], dtype=np.float32),)

    def backward(self, inputs, grad_output):
        return (self.grad,)


class ConstantFunction(chainer.Function):
    def __init__(self, val):
        super(ConstantFunction, self).__init__()
        self.val = val

    def forward(self, inputs):
        return (self.val,)

    def backward(self, inputs, grad_output):
        return ()


#
#     def get_sampler_reinf(self, fb_concat, mask, eos_idx, nb_steps = 50, use_best_for_sample = False,
#                     temperature = None,
#                     lexicon_probability_matrix = None,
#                     lex_epsilon = 1e-3,
#                     mode = "test"):
#         mb_size, nb_elems, Hi = fb_concat.data.shape
#         assert Hi == self.Hi, "%i != %i"%(Hi, self.Hi)
#
#         compute_ctxt = self.attn_module(fb_concat, mask)
#         while 1:
#             yield self.sample_reinf(self, nb_steps, compute_ctxt, mb_size, eos_idx, best = use_best_for_sample,
#                      lexicon_probability_matrix = lexicon_probability_matrix, lex_epsilon = lex_epsilon,
#                      temperature = temperature,
#                      mode = mode)
#
#
#     def compute_loss(self, targets, compute_ctxt, raw_loss_info = False, keep_attn_values = False,
#                      noise_on_prev_word = False, use_previous_prediction = 0, mode = "test", per_sentence = False,
#                      lexicon_probability_matrix = None, lex_epsilon = 1e-3
#                      ):
#         assert mode in "test train".split()
#
#         loss = None
#         current_mb_size = targets[0].data.shape[0]
#
#         previous_states = self.gru.get_initial_states(current_mb_size)
#
#         previous_word = None
#         with cuda.get_device(previous_states[0].data):
#             prev_y = F.broadcast_to(self.bos_embeding, (current_mb_size, self.Eo))
#         attn_list = []
#         total_nb_predictions = 0
#
#         if noise_on_prev_word:
#             noise_mean = Variable(self.xp.ones_like(prev_y.data, dtype = self.xp.float32))
#             noise_lnvar = Variable(self.xp.zeros_like(prev_y.data, dtype = self.xp.float32))
#
#         for i in xrange(len(targets)):
#             assert i == 0 or previous_states[0].data.shape[0] == previous_word.data.shape[0]
#             current_mb_size = targets[i].data.shape[0]
#             if current_mb_size < len(previous_states[0].data):
#                 truncated_states = [None] * len(previous_states)
#                 for num_state in xrange(len(previous_states)):
#                     truncated_states[num_state], _ = F.split_axis(previous_states[num_state], (current_mb_size,), 0)
#                 previous_states = tuple(truncated_states)
#
#                 if previous_word is not None:
#                     previous_word, _ = F.split_axis(previous_word, (current_mb_size,), 0 )
#
#                 if noise_on_prev_word:
#                     noise_mean, _ = F.split_axis(noise_mean, (current_mb_size,), 0)
#                     noise_lnvar, _ = F.split_axis(noise_lnvar, (current_mb_size,), 0)
#
#                 if lexicon_probability_matrix is not None:
#                     lexicon_probability_matrix = lexicon_probability_matrix[:current_mb_size]
#
#             if previous_word is not None: #else we are using the initial prev_y
#                 prev_y = self.emb(previous_word)
#             assert previous_states[0].data.shape[0] == current_mb_size
#
#             if noise_on_prev_word:
#                 prev_y = prev_y * F.gaussian(noise_mean, noise_lnvar)
#
#
#             new_states, logits, attn = self.advance_one_step(previous_states, prev_y,
#                                                     compute_ctxt, i, mode = mode)
#
#
#             if lexicon_probability_matrix is not None:
#                 # Just making sure data shape is as expected
#                 attn_mb_size, max_source_length_attn = attn.data.shape

#                 assert attn_mb_size == current_mb_size
#                 lex_mb_size, max_source_length_lexicon, v_size_lexicon = lexicon_probability_matrix.shape
#                 assert max_source_length_lexicon == max_source_length_attn
#                 assert logits.data.shape == (current_mb_size, v_size_lexicon)
#                 assert lex_mb_size == current_mb_size
#
# #                 weighted_lex_probs = F.reshape(
# #                         F.batch_matmul(attn, ConstantFunction(lexicon_probability_matrix)(), transa = True),
# #                                                logits.data.shape)
#
#                 weighted_lex_probs = F.reshape(
#                         batch_matmul_constant(attn, lexicon_probability_matrix, transa = True),
#                                                logits.data.shape)
#
#                 logits += F.log(weighted_lex_probs + lex_epsilon)
#
#
#
#             if keep_attn_values:
#                 attn_list.append(attn)
#
#             if per_sentence:
#                 normalized_logits = F.log_softmax(logits)
#                 total_local_loss = F.select_item(normalized_logits, targets[i])
#                 if loss is not None and total_local_loss.data.shape[0] != loss.data.shape[0]:
#                     assert total_local_loss.data.shape[0] < loss.data.shape[0]
#                     total_local_loss = F.concat(
#                                 (total_local_loss, self.xp.zeros(loss.data.shape[0] - total_local_loss.data.shape[0], dtype = self.xp.float32)),
#                                 axis = 0)
#             else:
#                 local_loss = F.softmax_cross_entropy(logits, targets[i])
#
#                 total_nb_predictions += current_mb_size
#                 total_local_loss = local_loss * current_mb_size
#
# #             loss = local_loss if loss is None else loss + local_loss#         assert loss_format in "average raw per_sentence".split(" ")
#             loss = total_local_loss if loss is None else loss + total_local_loss
#             if use_previous_prediction > 0 and random.random() < use_previous_prediction:
#                 previous_word = Variable(self.xp.argmax(logits.data, axis = 1).astype(self.xp.int32), volatile = "auto")
#             else:
#                 previous_word = targets[i]
# #             prev_y = self.emb(previous_word)
#             previous_states = new_states
# #             attn_list.append(attn)
#         if raw_loss_info:
#             return (loss, total_nb_predictions), attn_list
#         else:
#             loss = loss / total_nb_predictions
#             return loss, attn_list
#
#     def compute_loss_and_backward(self, fb_concat, targets, mask,
#                      noise_on_prev_word = False, use_previous_prediction = 0):
#
#         compute_ctxt = self.attn_module(fb_concat, mask)
#
#         loss = 0
#         pseudo_loss = 0
#
#         current_mb_size = targets[0].data.shape[0]
#
#         previous_states = self.gru.get_initial_states(current_mb_size)
#
#         previous_word = None
#         with cuda.get_device(previous_states[0].data):
#             prev_y = F.broadcast_to(self.bos_embeding, (current_mb_size, self.Eo))
#
#         if noise_on_prev_word:
#             noise_mean = Variable(self.xp.ones_like(prev_y.data, dtype = self.xp.float32))
#             noise_lnvar = Variable(self.xp.zeros_like(prev_y.data, dtype = self.xp.float32))
#
#         total_nb_predictions = sum(t.data.shape[0] for t in targets)
#
#         for i in xrange(len(targets)):
#             assert i == 0 or previous_states[0].data.shape[0] == previous_word.data.shape[0]
#             current_mb_size = targets[i].data.shape[0]
#             if current_mb_size < len(previous_states[0].data):
#                 truncated_states = [None] * len(previous_states)
#                 for num_state in xrange(len(previous_states)):
#                     truncated_states[num_state], _ = F.split_axis(previous_states[num_state], (current_mb_size,), 0)
#                 previous_states = tuple(truncated_states)
#
#                 if previous_word is not None:
#                     previous_word, _ = F.split_axis(previous_word, (current_mb_size,), 0 )
#
#                 if noise_on_prev_word:
#                     noise_mean, _ = F.split_axis(noise_mean, (current_mb_size,), 0)
#                     noise_lnvar, _ = F.split_axis(noise_lnvar, (current_mb_size,), 0)
#             if previous_word is not None: #else we are using the initial prev_y
#                 prev_y = self.emb(previous_word)
#             assert previous_states[0].data.shape[0] == current_mb_size
#
#             if noise_on_prev_word:
#                 prev_y = prev_y * F.gaussian(noise_mean, noise_lnvar)
#
#
#             output_state = previous_states[-1]
#             ci, attn = compute_ctxt(output_state)
#             concatenated = F.concat( (prev_y, ci) )
#
#             new_states = self.gru(previous_states, concatenated, mode = "train")
#             new_output_state = new_states[-1]
#
#             all_concatenated = F.concat((concatenated, new_output_state))
#
#             all_concatenated_bis = Variable(all_concatenated.data)
#             logits = self.lin_o(self.maxo(all_concatenated_bis))
#
#             partial_loss = F.softmax_cross_entropy(logits, targets[i]) * current_mb_size / total_nb_predictions
#             loss += partial_loss.data
#             partial_loss.backward()
#             gr = GradKeeper(all_concatenated_bis.grad)
#             del logits
#             del all_concatenated_bis
#             del partial_loss
#
#             pseudo_loss += gr(all_concatenated)
#
#             if use_previous_prediction > 0 and random.random() < use_previous_prediction:
#                 previous_word = Variable(self.xp.argmax(logits.data, axis = 1).astype(self.xp.int32), volatile = "auto")
#             else:
#                 previous_word = targets[i]
#
#             previous_states = new_states
#
#         pseudo_loss.backward()
#         return loss, total_nb_predictions
#
#
#
# #
#     def get_predictor(self, fb_concat, mask):
#         mb_size, nb_elems, Hi = fb_concat.data.shape
#         assert Hi == self.Hi, "%i != %i"%(Hi, self.Hi)
#         xp = cuda.get_array_module(self.initial_state.data)
#
#         compute_ctxt = self.attn_module.compute_ctxt_demux(fb_concat, mask)
#
#         assert mb_size == 1
#         current_mb_size = mb_size
# #         previous_state = F.concat( [self.initial_state] * current_mb_size, 0)
#         previous_state = [F.broadcast_to(self.initial_state, (current_mb_size, self.Ho))]
# #         previous_word = Variable(np.array([self.bos_idx] * mb_size, dtype = np.int32))
#         previous_word = [None]
#         with cuda.get_device(self.initial_state.data):
# #             previous_word = Variable(xp.array([self.bos_idx] * current_mb_size, dtype = np.int32))
#             prev_y_initial = F.broadcast_to(self.bos_embeding, (current_mb_size, self.Eo))
#
#
#         def choose(voc_list, i):
#             if previous_word[0] is not None: #else we are using the initial prev_y
#                 prev_y = self.emb(previous_word[0])
#             else:
#                 prev_y = prev_y_initial
#             assert previous_state[0].data.shape[0] == current_mb_size
#
#             new_state, logits, attn = self.advance_one_step(previous_state[0], prev_y,
#                                                       compute_ctxt, i, mode = "test")
#
#             best_w = None
#             best_score = None
#             for w in voc_list:
#                 score = logits.data[0][w]
#                 if best_score is None or score > best_score:
#                     best_score = score
#                     best_w = w
#
#
#             previous_word[0] = Variable(self.xp.array([best_w], dtype = self.xp.int32), volatile = "auto")
#             previous_state[0] = new_state
#             return best_w
#         return choose
#
#
#     def __call__(self, fb_concat, targets, mask, use_best_for_sample = False, raw_loss_info = False,
#                     keep_attn_values = False, need_score = False, noise_on_prev_word = False,
#                     use_previous_prediction = 0, mode = "test", lexicon_probability_matrix = None, lex_epsilon = 1e-3):
#         assert mode in "test train".split()
#         mb_size, nb_elems, Hi = fb_concat.data.shape
#         assert Hi == self.Hi, "%i != %i"%(Hi, self.Hi)
#
#         compute_ctxt = self.attn_module(fb_concat, mask)
#
#         if isinstance(targets, int):
#             return self.sample(targets, compute_ctxt, mb_size, best = use_best_for_sample,
#                                keep_attn_values = keep_attn_values, need_score = need_score,
#                                lexicon_probability_matrix = lexicon_probability_matrix, lex_epsilon = lex_epsilon)
#         else:
#             return self.compute_loss(targets, compute_ctxt, raw_loss_info = raw_loss_info,
#                                      keep_attn_values = keep_attn_values, noise_on_prev_word = noise_on_prev_word,
#                                      use_previous_prediction = use_previous_prediction, mode = mode,
#                                      lexicon_probability_matrix = lexicon_probability_matrix, lex_epsilon = lex_epsilon)
#
# class EncoderDecoder(Chain):
#     """ Do RNNSearch Encoding/Decoding
#         The __call__ takes 3 required parameters: src_batch, tgt_batch, src_mask
#         src_batch is as in the sequence parameter of Encoder
#         tgt_batch is as in the targets parameter of Decoder
#         src_mask is as in the mask parameter of Encoder
#
#         return loss and attention values
#     """
#     def __init__(self, Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl, attn_cls = AttentionModule, init_orth = False, use_bn_length = 0,
#                 encoder_cell_type = "gru", decoder_cell_type = "gru",
#                 lexical_probability_dictionary = None, lex_epsilon = 1e-3
#                 ):
#         log.info("constructing encoder decoder with Vi:%i Ei:%i Hi:%i Vo:%i Eo:%i Ho:%i Ha:%i Hl:%i" %
#                                         (Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl))
#         super(EncoderDecoder, self).__init__(
#             enc = Encoder(Vi, Ei, Hi, init_orth = init_orth, use_bn_length = use_bn_length,
#                           cell_type = encoder_cell_type),
#             dec = Decoder(Vo, Eo, Ho, Ha, 2 * Hi, Hl, attn_cls = attn_cls, init_orth = init_orth,
#                           use_bn_length = use_bn_length, cell_type = decoder_cell_type)
#         )
#         self.Vo = Vo
#         self.lexical_probability_dictionary = lexical_probability_dictionary
#         self.lex_epsilon = lex_epsilon
#
#     def __call__(self, src_batch, tgt_batch, src_mask, use_best_for_sample = False, display_attn = False,
#                  raw_loss_info = False, keep_attn_values = False, need_score = False, noise_on_prev_word = False,
#                  use_previous_prediction = 0, mode = "test"
#                  ):
#         assert mode in "test train".split()
#
#         if self.lexical_probability_dictionary is not None:
#             lexicon_probability_matrix = compute_lexicon_matrix(src_batch, self.lexical_probability_dictionary, self.Vo)
#             if self.xp != np:
#                 lexicon_probability_matrix = cuda.to_gpu(lexicon_probability_matrix, cuda.get_device(self.dec.lin_o.W.data))
#         else:
#             lexicon_probability_matrix = None
#
#         fb_src = self.enc(src_batch, src_mask, mode = mode)
#         loss = self.dec(fb_src, tgt_batch, src_mask, use_best_for_sample = use_best_for_sample, raw_loss_info = raw_loss_info,
#                         keep_attn_values = keep_attn_values, need_score = need_score, noise_on_prev_word = noise_on_prev_word,
#                         mode = mode, use_previous_prediction = use_previous_prediction,
#                         lexicon_probability_matrix = lexicon_probability_matrix, lex_epsilon = self.lex_epsilon)
#         return loss
#
# #     def compute_loss_and_backward(self, src_batch, tgt_batch, src_mask, raw_loss_info = False):
# #         fb_src = self.enc(src_batch, src_mask)
# #         loss = self.dec.compute_loss_and_backward(fb_src, tgt_batch, src_mask, raw_loss_info = raw_loss_info)
# #         return loss
#
#     def compute_loss_and_backward(self, src_batch, tgt_batch, src_mask,
#                      noise_on_prev_word = False, use_previous_prediction = 0):
#
#         fb_concat = self.enc(src_batch, src_mask, mode = "train")
#         return self.dec.compute_loss_and_backward(fb_concat, tgt_batch, src_mask,
#                      noise_on_prev_word = False, use_previous_prediction = 0)
#
# #     def sample(self, src_batch, src_mask, nb_steps, use_best_for_sample, keep_attn_values = False, need_score = False):
# #         fb_src = self.enc(src_batch, src_mask)
# #         samp = self.dec.sample(self, fb_src, nb_steps, src_mask, use_best_for_sample = use_best_for_sample,
# #                         keep_attn_values = keep_attn_values, need_score = need_score)
# #         return samp
# #         return self.dec.beam_search(fb_src, src_mask, nb_steps, eos_idx = eos_idx, beam_width = beam_width)
#
#
#     def get_predictor(self, src_batch, src_mask):
#         fb_src = self.enc(src_batch, src_mask)
#         return self.dec.get_predictor(fb_src, src_mask)
#
#     def nbest_scorer(self, src_batch, src_mask):
#         assert len(src_batch[0].data) == 1
#         fb_concat = self.enc(src_batch, src_mask)
#         compute_ctxt = self.dec.attn_module.compute_ctxt_demux(fb_concat, src_mask)
#
#         def scorer(tgt_batch):
#             return self.dec.compute_loss(tgt_batch, compute_ctxt, raw_loss_info = True, keep_attn_values = False,
#                  noise_on_prev_word = False, use_previous_prediction = 0, mode = "test", per_sentence = True)
#
#         return scorer
#
#     def get_sampler_reinf(self, src_batch, src_mask, eos_idx, nb_steps = 50, use_best_for_sample = False,
#                     temperature = None,
#                     mode = "test"):
#         fb_concat = self.enc(src_batch, src_mask)
#         return self.dec.get_sampler_reinf(fb_concat, src_mask, eos_idx, nb_steps = nb_steps,
#                     use_best_for_sample = use_best_for_sample,
#                     temperature = temperature,
#                     lexicon_probability_matrix = self.lexicon_probability_matrix,
#                     lex_epsilon = self.lex_epsilon,
#                     mode = mode)
#
#     def get_reinf_loss(self, src_batch, src_mask, eos_idx, test_references, nb_steps = 50, nb_samples = 5, use_best_for_sample = False,
#                     temperature = None,
#                     mode = "test"):
#
#         mb_size = len(src_batch[0])
#         sampler = self.get_sampler_reinf(self, src_batch, src_mask, eos_idx, nb_steps = nb_steps,
#                     use_best_for_sample = use_best_for_sample,
#                     temperature = None,
#                     mode = mode)
#
#         from utils import de_batch
#         from bleu_computer import BleuComputer
#         total_score = 0
#         for i in xrange(nb_samples):
#             sentences, score = sampler.next()
#
#             deb = de_batch(sentences, mask = None, eos_idx = eos_idx, is_variable = False)
#
#             assert len(deb) == len(test_references)
#             assert len(deb) == mb_size
#
#             bleu_vec = np.zeros((mb_size, 1), dtype = np.float32)
#             for num_t in range(len(deb)):
#                 t = deb[num_t]
#                 if t[-1] == eos_idx:
#                     t = t[:-1]
#                 ref = test_references[num_t]
#                 bc = BleuComputer()
#                 bc.update(ref, t)
#                 bleu = bc.bleu_plus_alpha(0.5)
#                 bleu_vec[num_t] = bleu
#
#             if self.xp is not np:
#                 bleu_vec = cuda.to_gpu(bleu_vec)
#
#             weighted_score = F.matmul(score, Variable(bleu_vec, volatile = "auto"))
#             total_score += weighted_score
#
#         total_score /= nb_samples
#         return total_score

import decoder_cells
import encoders


class EncoderDecoder(Chain):
    """ Do RNNSearch Encoding/Decoding

        Args:
            Vi: Source vocabulary size
            Ei: Size of Source word embeddings
            Hi: Size of source encoder hidden layer
            Vo: Target Vocabulary Size
            Eo: Size of Target word embeddings
            Ho: Size of Decoder hidden state
            Ha: Size of Attention mechanism hidden layer size
            Hl: Size of maxout output layer
            attn_cls: class of the Attention to be used


        The __call__ takes 3 required parameters: src_batch, tgt_batch, src_mask
        src_batch is as in the sequence parameter of Encoder
        tgt_batch is as in the targets parameter of Decoder
        src_mask is as in the mask parameter of Encoder

        return loss and attention values
    """

    def __init__(self, Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl, attn_cls=attention.AttentionModule, init_orth=False, use_bn_length=0,
                 encoder_cell_type=rnn_cells.LSTMCell,
                 decoder_cell_type=rnn_cells.LSTMCell,
                 lexical_probability_dictionary=None, lex_epsilon=1e-3
                 ):
        log.info("constructing encoder decoder with Vi:%i Ei:%i Hi:%i Vo:%i Eo:%i Ho:%i Ha:%i Hl:%i" %
                 (Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl))
        super(EncoderDecoder, self).__init__(
            enc=encoders.make_encoder(Vi, Ei, Hi, init_orth=init_orth, use_bn_length=use_bn_length,
                                      cell_type=encoder_cell_type),
            dec=decoder_cells.Decoder(Vo, Eo, Ho, Ha, 2 * Hi, Hl, attn_cls=attn_cls, init_orth=init_orth,
                                      cell_type=decoder_cell_type)
        )
        self.Vo = Vo
        self.lexical_probability_dictionary = lexical_probability_dictionary
        self.lex_epsilon = lex_epsilon

    def compute_lexicon_probability_matrix(self, src_batch):
        if self.lexical_probability_dictionary is not None:
            lexicon_probability_matrix = compute_lexicon_matrix(
                src_batch, self.lexical_probability_dictionary, self.Vo)
            if self.xp != np:
                lexicon_probability_matrix = cuda.to_gpu(
                    lexicon_probability_matrix, cuda.get_device(
                        self.dec.lin_o.W.data))
        else:
            lexicon_probability_matrix = None
        return lexicon_probability_matrix

    def __call__(self, src_batch, tgt_batch, src_mask, use_best_for_sample=False, display_attn=False,
                 raw_loss_info=False, keep_attn_values=False, need_score=False, noise_on_prev_word=False,
                 use_previous_prediction=0, mode="test"
                 ):
        assert mode in "test train".split()

        lexicon_probability_matrix = self.compute_lexicon_probability_matrix(src_batch)

        fb_concat = self.enc(src_batch, src_mask, mode=mode)

        mb_size, nb_elems, Hi = fb_concat.data.shape

        if isinstance(tgt_batch, int):
            return self.dec.sample(fb_concat, src_mask, tgt_batch, mb_size,
                                   lexicon_probability_matrix=lexicon_probability_matrix,
                                   lex_epsilon=self.lex_epsilon, best=use_best_for_sample,
                                   keep_attn_values=keep_attn_values, need_score=need_score)

        else:
            return self.dec.compute_loss(fb_concat, src_mask, tgt_batch, raw_loss_info=raw_loss_info,
                                         keep_attn_values=keep_attn_values, noise_on_prev_word=noise_on_prev_word,
                                         use_previous_prediction=use_previous_prediction, mode="test",
                                         per_sentence=False, lexicon_probability_matrix=lexicon_probability_matrix,
                                         lex_epsilon=self.lex_epsilon)

    def give_conditionalized_cell(self, src_batch, src_mask, noise_on_prev_word=False,
                                  mode="test", demux=False):

        if self.lexical_probability_dictionary is not None:
            lexicon_probability_matrix = compute_lexicon_matrix(
                src_batch, self.lexical_probability_dictionary, self.Vo)
            if self.xp != np:
                lexicon_probability_matrix = cuda.to_gpu(
                    lexicon_probability_matrix, cuda.get_device(
                        self.dec.lin_o.W.data))
        else:
            lexicon_probability_matrix = None

        fb_concat = self.enc(src_batch, src_mask, mode=mode)

        mb_size, nb_elems, Hi = fb_concat.data.shape

        return self.dec.give_conditionalized_cell(fb_concat, src_mask,
                                                  noise_on_prev_word=noise_on_prev_word,
                                                  mode=mode, lexicon_probability_matrix=lexicon_probability_matrix,
                                                  lex_epsilon=self.lex_epsilon, demux=demux)

    def nbest_scorer(self, src_batch, src_mask, keep_attn=False):
        assert len(src_batch[0].data) == 1

        lexicon_probability_matrix = self.compute_lexicon_probability_matrix(src_batch)
        fb_concat = self.enc(src_batch, src_mask)

        decoding_cell = self.dec.give_conditionalized_cell(fb_concat, src_mask, noise_on_prev_word=False,
                                                           mode="test", lexicon_probability_matrix=lexicon_probability_matrix, lex_epsilon=self.lex_epsilon,
                                                           demux=True)

        def scorer(tgt_batch):

            loss, attn_list = decoder_cells.compute_loss_from_decoder_cell(decoding_cell, tgt_batch,
                                                                           use_previous_prediction=0,
                                                                           raw_loss_info=True,
                                                                           per_sentence=True,
                                                                           keep_attn=keep_attn)

            return loss, attn_list

        return scorer

#     def compute_loss_and_backward(self, src_batch, tgt_batch, src_mask, raw_loss_info = False):
#         fb_src = self.enc(src_batch, src_mask)
#         loss = self.dec.compute_loss_and_backward(fb_src, tgt_batch, src_mask, raw_loss_info = raw_loss_info)
#         return loss


# class AttentionModuleAcumulated(Chain):
#     """ Attention Module for computing the current context during decoding.
#         The __call_ takes 2 parameters: fb_concat and mask.
#
#         fb_concat should be the result of a call to Encoder.
#         mask is as in the description of Encoder
#
#         Return a chainer variable of shape (mb_size, Hi) and type float32
#     """
#     def __init__(self, Hi, Ha, Ho):
#         super(AttentionModuleAcumulated, self).__init__(
#             al_lin_h = L.Linear(Hi, Ha, nobias = False),
#             al_lin_s = L.Linear(Ho, Ha, nobias = True),
#             al_lin_o = L.Linear(Ha, 1, nobias = True)
#         )
#         self.Hi = Hi
#         self.Ha = Ha
#
#     def __call__(self, fb_concat, mask):
#         mb_size, nb_elems, Hi = fb_concat.data.shape
#         assert Hi == self.Hi
#         precomputed_al_factor = F.reshape(self.al_lin_h(
#                         F.reshape(fb_concat, (mb_size * nb_elems, self.Hi)))
#                                           , (mb_size, nb_elems, self.Ha))
#
#         concatenated_mask = F.concat([F.reshape(mask_elem, (mb_size, 1)) for mask_elem in mask], 1)
#
#         accumulation = [Variable(self.xp.zeros((mb_size, nb_elems), dtype = self.xp.float32), volatile = "auto")]
#
#         def compute_ctxt(previous_state):
#             current_mb_size = previous_state.data.shape[0]
#             if current_mb_size < mb_size:
#                 al_factor, _ = F.split_axis(precomputed_al_factor, (current_mb_size,), 0)
#                 used_fb_concat, _ = F.split_axis(fb_concat, (current_mb_size,), 0)
#                 used_concatenated_mask, _ = F.split_axis(concatenated_mask, (current_mb_size,), 0)
#
#                 # Warning: here we are actually changing the global accumulation
#                 if current_mb_size < accumulation[0].data.shape[0]:
#                     accumulation[0], _ = F.split_axis(accumulation[0], (current_mb_size,), 0)
#             else:
#                 al_factor = precomputed_al_factor
#                 used_fb_concat = fb_concat
#                 used_concatenated_mask = concatenated_mask
#
#             state_al_factor = self.al_lin_s(previous_state)
#             state_al_factor_bc = F.broadcast_to(F.reshape(state_al_factor, (current_mb_size, 1, self.Ha)), (current_mb_size, nb_elems, self.Ha) )
#             a_coeffs = F.reshape(self.al_lin_o(F.reshape(F.tanh(state_al_factor_bc + al_factor),
#                             (current_mb_size* nb_elems, self.Ha))), (current_mb_size, nb_elems))
#
#             a_coeffs_mods = a_coeffs -  accumulation[0]
#             accumulation[0] += a_coeffs
#
#             with cuda.get_device(used_concatenated_mask.data):
#                 a_coeffs_mods = a_coeffs_mods - 50000 * (1-used_concatenated_mask.data)
#
#             attn = F.softmax(a_coeffs_mods)
#
#             ci = F.reshape(F.batch_matmul(attn, used_fb_concat, transa = True), (current_mb_size, self.Hi))
#
#             return ci, attn
#
#         return compute_ctxt
#
#     def compute_ctxt_demux(self, fb_concat, mask):
#         raise NotImplemented
#         mb_size, nb_elems, Hi = fb_concat.data.shape
#         assert Hi == self.Hi
#         assert mb_size == 1
#
#         precomputed_al_factor = F.reshape(self.al_lin_h(
#                         F.reshape(fb_concat, (mb_size * nb_elems, self.Hi)))
#                                           , (mb_size, nb_elems, self.Ha))
#
#         concatenated_mask = F.concat([F.reshape(mask_elem, (mb_size, 1)) for mask_elem in mask], 1)
#
#
#         def compute_ctxt(previous_state):
#             current_mb_size = previous_state.data.shape[0]
#
#             al_factor = F.broadcast_to(precomputed_al_factor, (current_mb_size, nb_elems, self.Ha))
# #             used_fb_concat = F.broadcast_to(fb_concat, (current_mb_size, nb_elems, Hi))
#             used_concatenated_mask = F.broadcast_to(concatenated_mask, (current_mb_size, nb_elems))
#
#             state_al_factor = self.al_lin_s(previous_state)
#             state_al_factor_bc = F.broadcast_to(F.reshape(state_al_factor, (current_mb_size, 1, self.Ha)), (current_mb_size, nb_elems, self.Ha) )
#             a_coeffs = F.reshape(self.al_lin_o(F.reshape(F.tanh(state_al_factor_bc + al_factor),
#                             (current_mb_size* nb_elems, self.Ha))), (current_mb_size, nb_elems))
#
#
#             with cuda.get_device(used_concatenated_mask.data):
#                 a_coeffs = a_coeffs - 10000 * (1-used_concatenated_mask.data)
#
#             attn = F.softmax(a_coeffs)
#
# #             ci = F.reshape(F.batch_matmul(attn, used_fb_concat, transa = True), (current_mb_size, self.Hi))
#
#             ci = F.reshape(F.matmul(attn, F.reshape(fb_concat, (nb_elems, Hi))), (current_mb_size, self.Hi))
#
#             return ci, attn
#
#         return compute_ctxt
#
# class DecoderWithAlign(Chain):
#     """ Decoder for RNNSearch.
#         The __call_ takes 3 required parameters: fb_concat, targets, mask.
#
#         fb_concat should be the result of a call to Encoder.
#
#         targets is a python list of chainer variables of type int32 and of variable shape (n,)
#             the values n should be decreasing:
#                 i < j => targets[i].data.shape[0] >= targets[j].data.shape[0]
#             targets[i].data[j] is the jth elements of the ith sequence in the minibatch
#             all this imply that the sequences of the minibatch should be sorted from longest to shortest
#
#         mask is as in the description of Encoder.
#
#         * it is up to the user to add an EOS token to the data.
#
#         Return a loss and the attention model values
#     """
#     def __init__(self, Vo, Eo, Ho, Ha, Hi, Hl, attn_cls = AttentionModule, init_orth = False):
#         super(DecoderWithAlign, self).__init__(
#             emb = L.EmbedID(Vo, Eo),
#             gru = L.GRU(Ho, Eo + Hi),
#
#             maxo = L.Maxout(Eo + Hi + Ho, Hl, 2),
#             lin_o = L.Linear(Hl, Vo, nobias = False),
#
#             attn_module = attn_cls(Hi, Ha, Ho, init_orth = init_orth),
#
#             align_pred_mo = L.Maxout(Ho, 1,4),
#             size_pred_mo = L.Maxout(Hi, 2, 2)
#         )
#         self.add_param("initial_state", (1, Ho))
#         self.add_param("bos_embeding", (1, Eo))
#         self.Hi = Hi
#         self.Ho = Ho
#         self.Eo = Eo
#         self.initial_state.data[...] = np.random.randn(Ho)
#         self.bos_embeding.data[...] = np.random.randn(Eo)
#
#         if init_orth:
#             ortho_init(self.gru)
#             ortho_init(self.lin_o)
#             ortho_init(self.maxo)
#
#     def advance_one_step(self, previous_state, prev_y, compute_ctxt):
#
#         ci, attn = compute_ctxt(previous_state)
#         concatenated = F.concat( (prev_y, ci) )
# #             print concatenated.data.shape
#         new_state = self.gru(previous_state, concatenated)
#
#         all_concatenated = F.concat((concatenated, new_state))
#         logits = self.lin_o(self.maxo(all_concatenated))
#
#         al_pred =  self.align_pred_mo(new_state)
#
#         return new_state, logits, attn, al_pred
#
#     def compute_loss(self, targets, fb_concat, mask, sizes, raw_loss_info = False, keep_attn_values = False, noise_on_prev_word = False):
#         compute_ctxt = self.attn_module(fb_concat, mask)
#         loss = None
#         current_mb_size = targets[0].data.shape[0]
# #         previous_state = F.concat( [self.initial_state] * current_mb_size, 0)
#         previous_state = F.broadcast_to(self.initial_state, (current_mb_size, self.Ho))
# #         previous_word = Variable(np.array([self.bos_idx] * mb_size, dtype = np.int32))
#         xp = cuda.get_array_module(self.initial_state.data)
#         previous_word = None
#         with cuda.get_device(self.initial_state.data):
# #             previous_word = Variable(xp.array([self.bos_idx] * current_mb_size, dtype = np.int32))
#             prev_y = F.broadcast_to(self.bos_embeding, (current_mb_size, self.Eo))
#         attn_list = []
#         total_nb_predictions = 0
#
#         if noise_on_prev_word:
#             noise_mean = Variable(self.xp.ones_like(prev_y.data, dtype = self.xp.float32))
#             noise_lnvar = Variable(self.xp.zeros_like(prev_y.data, dtype = self.xp.float32))
#
#         first_fb, _ = F.split_axis(fb_concat, (1,), 1)
#         size_pred = F.softplus(self.size_pred_mo(first_fb))
#         size_mean, size_precision = F.split_axis(size_pred, (1,), 1)
#         loss_size = 0.5 * (size_precision* (sizes - size_mean) ) **2 - F.log(size_precision)
#         loss = F.sum(loss_size)/ current_mb_size
#
#         for i in xrange(len(targets)):
#             assert i == 0 or previous_state.data.shape[0] == previous_word.data.shape[0]
#             current_mb_size = targets[i].data.shape[0]
#             if current_mb_size < len(previous_state.data):
#                 previous_state, _ = F.split_axis(previous_state, (current_mb_size,), 0)
#                 if previous_word is not None:
#                     previous_word, _ = F.split_axis(previous_word, (current_mb_size,), 0 )
#
#                 if noise_on_prev_word:
#                     noise_mean, _ = F.split_axis(noise_mean, (current_mb_size,), 0)
#                     noise_lnvar, _ = F.split_axis(noise_lnvar, (current_mb_size,), 0)
#             if previous_word is not None: #else we are using the initial prev_y
#                 prev_y = self.emb(previous_word)
#             assert previous_state.data.shape[0] == current_mb_size
#
#             if noise_on_prev_word:
#                 prev_y = prev_y * F.gaussian(noise_mean, noise_lnvar)
#
#             new_state, logits, attn, al_pred = self.advance_one_step(previous_state, prev_y,
#                                                       compute_ctxt)
#
#             if keep_attn_values:
#                 attn_list.append(attn)
#
#             local_loss = F.softmax_cross_entropy(logits, targets[i][0])
#
#             local_loss += F.softmax_cross_entropy(al_pred, targets[i][1])
#
#             total_nb_predictions += current_mb_size
#             total_local_loss = local_loss * current_mb_size
#
# #             loss = local_loss if loss is None else loss + local_loss
#             loss = total_local_loss if loss is None else loss + total_local_loss
#
#             previous_word = targets[i]
# #             prev_y = self.emb(previous_word)
#             previous_state = new_state
# #             attn_list.append(attn)
#         if raw_loss_info:
#             return (loss, total_nb_predictions), attn_list
#         else:
#             loss = loss / total_nb_predictions
#             return loss, attn_list

# class EncoderDecoderPredictAlign(Chain):
#     """ Do RNNSearch Encoding/Decoding
#         The __call__ takes 3 required parameters: src_batch, tgt_batch, src_mask
#         src_batch is as in the sequence parameter of Encoder
#         tgt_batch is as in the targets parameter of Decoder
#         src_mask is as in the mask parameter of Encoder
#
#         return loss and attention values
#     """
#     def __init__(self, Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl, attn_cls = AttentionModule, init_orth = False):
#         log.info("constructing encoder decoder with Vi:%i Ei:%i Hi:%i Vo:%i Eo:%i Ho:%i Ha:%i Hl:%i" %
#                                         (Vi, Ei, Hi, Vo, Eo, Ho, Ha, Hl))
#         super(EncoderDecoderPredictAlign, self).__init__(
#             enc = Encoder(Vi, Ei, Hi, init_orth = init_orth),
#             dec = DecoderWithAlign(Vo, Eo, Ho, Ha, 2 * Hi, Hl, attn_cls = attn_cls, init_orth = init_orth)
#         )
#
#     def __call__(self, src_batch, tgt_batch, src_mask, sizes, use_best_for_sample = False, display_attn = False,
#                  raw_loss_info = False, keep_attn_values = False, need_score = False, noise_on_prev_word = False):
#         fb_src = self.enc(src_batch, src_mask)
#         loss = self.dec.compute_loss(tgt_batch, fb_src, src_mask, sizes, use_best_for_sample = use_best_for_sample, raw_loss_info = raw_loss_info,
#                         keep_attn_values = keep_attn_values, need_score = need_score, noise_on_prev_word = noise_on_prev_word)
#         return loss



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""make_data.py: prepare data for training"""
import nmt_chainer
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import collections
import logging
import codecs
import json
import operator
import os.path
import gzip

from nmt_chainer.utilities.utils import ensure_path
import nmt_chainer.dataprocessing.processors as processors

logging.basicConfig()
log = logging.getLogger("rnns:make_data")
log.setLevel(logging.INFO)


def do_make_data(config):
    #     raw_input("Press Enter to Continue 222")

    save_prefix_dir, save_prefix_fn = os.path.split(config.data.save_prefix)
    ensure_path(save_prefix_dir)

    config_fn = config.data.save_prefix + ".data.config"
    voc_fn = config.data.save_prefix + ".voc"
    data_fn = config.data.save_prefix + ".data.json.gz"
#     valid_data_fn = config.save_prefix + "." + config.model + ".valid.data.npz"

#     voc_fn_src = config.save_prefix + ".src.voc"
#     voc_fn_tgt = config.save_prefix + ".tgt.voc"

    files_that_will_be_created = [config_fn, voc_fn, data_fn]

    if config.processing.bpe_src is not None:
        bpe_data_file_src = config.data.save_prefix + ".src.bpe"
        files_that_will_be_created.append(bpe_data_file_src)

    if config.processing.bpe_tgt is not None:
        bpe_data_file_tgt = config.data.save_prefix + ".tgt.bpe"
        files_that_will_be_created.append(bpe_data_file_tgt)

    if config.processing.joint_bpe is not None:
        bpe_data_file_joint = config.data.save_prefix + ".joint.bpe"
        files_that_will_be_created.append(bpe_data_file_joint)

    already_existing_files = []
    for filename in files_that_will_be_created:  # , valid_data_fn]:
        if os.path.exists(filename):
            already_existing_files.append(filename)
    if len(already_existing_files) > 0:
        print "Warning: existing files are going to be replaced: ", already_existing_files
        raw_input("Press Enter to Continue")

    if config.processing.use_voc is not None:
        log.info("loading voc from %s" % config.processing.use_voc)
#         src_voc, tgt_voc = json.load(open(config.use_voc))
#         src_pp = processors.load_pp_from_data(json.load(open(src_voc)))
#         tgt_pp = IndexingPrePostProcessor.make_from_serializable(tgt_voc)
        bi_idx = processors.load_pp_pair_from_file(config.processing.use_voc)
    else:

        bi_idx = processors.BiIndexingPrePostProcessor(voc_limit1=config.processing.src_voc_size,
                                                       voc_limit2=config.processing.tgt_voc_size)
        pp = processors.BiProcessorChain()

        if config.processing.latin_tgt:
            pp.add_tgt_processor(processors.LatinScriptProcess(config.processing.latin_type))

        if config.processing.latin_src:
            pp.add_src_processor(processors.LatinScriptProcess(config.processing.latin_type))

        pp.add_src_processor(processors.SimpleSegmenter(config.processing.src_segmentation_type))
        if config.processing.bpe_src is not None:
            pp.add_src_processor(
                processors.BPEProcessing(bpe_data_file=bpe_data_file_src, symbols=config.processing.bpe_src, separator="._@@@"))

        pp.add_tgt_processor(processors.SimpleSegmenter(config.processing.tgt_segmentation_type))
        if config.processing.bpe_tgt is not None:
            pp.add_tgt_processor(
                processors.BPEProcessing(bpe_data_file=bpe_data_file_tgt, symbols=config.processing.bpe_tgt, separator="._@@@"))

        if config.processing.joint_bpe is not None:
            pp.add_biprocessor(processors.JointBPEBiProcessor(bpe_data_file=bpe_data_file_joint,
                                                              symbols=config.processing.joint_bpe, separator="._@@@"))

        bi_idx.add_preprocessor(pp)

    def load_data(src_fn, tgt_fn, max_nb_ex=None, infos_dict=None):

        training_data, stats_src, stats_tgt = processors.build_dataset_pp(
            src_fn, tgt_fn, bi_idx,
            max_nb_ex=max_nb_ex)

        log.info("src data stats:\n%s", stats_src.make_report())
        log.info("tgt data stats:\n%s", stats_tgt.make_report())

        if infos_dict is not None:
            infos_dict["src"] = stats_src.report_as_obj()
            infos_dict["tgt"] = stats_tgt.report_as_obj()

        return training_data

    infos = collections.OrderedDict()
    infos["train"] = collections.OrderedDict()

    log.info("loading training data from %s and %s" %
             (config.data.src_fn, config.data.tgt_fn))
    training_data = load_data(config.data.src_fn, config.data.tgt_fn, max_nb_ex=config.data.max_nb_ex, infos_dict=infos["train"])

    dev_data = None
    if config.data.dev_src is not None:
        log.info("loading dev data from %s and %s" %
                 (config.data.dev_src, config.data.dev_tgt))
        infos["dev"] = collections.OrderedDict()
        dev_data = load_data(
            config.data.dev_src, config.data.dev_tgt, infos_dict=infos["dev"])

    test_data = None
    if config.data.test_src is not None:
        log.info("loading test data from %s and %s" %
                 (config.data.test_src, config.data.test_tgt))
        infos["test"] = collections.OrderedDict()
        test_data = load_data(
            config.data.test_src, config.data.test_tgt, infos_dict=infos["test"])

    config.insert_section("infos", infos, even_if_readonly=True, keep_at_bottom="metadata", overwrite=False)

#     if config.shuffle:
#         log.info("shuffling data")
#         if config.enable_fast_shuffle:
#             shuffle_in_unison_faster(data_input, data_target)
#         else:
#             data_input, data_target = shuffle_in_unison(data_input, data_target)
    log.info("saving config to %s" % config_fn)
    config.save_to(config_fn)
#     json.dump(config.__dict__, open(config_fn, "w"),
#               indent=2, separators=(',', ': '))

    log.info("saving voc to %s" % voc_fn)
    processors.save_pp_pair_to_file(bi_idx, voc_fn)
#     json.dump([src_pp.to_serializable(), tgt_pp.to_serializable()],
#               open(voc_fn, "w"), indent=2, separators=(',', ': '))

    log.info("saving train_data to %s" % data_fn)
    data_all = {"train": training_data}
    if test_data is not None:
        data_all["test"] = test_data
    if dev_data is not None:
        data_all["dev"] = dev_data

    json.dump(data_all, gzip.open(data_fn, "wb"),
              indent=2, separators=(',', ': '))



"""train_config.py: Parse training arguments and create config dictionnary."""

import logging
import sys

from nmt_chainer.utilities import argument_parsing_tools

logging.basicConfig()
log = logging.getLogger("rnns:train_config")
log.setLevel(logging.INFO)

_CONFIG_SECTION_TO_DESCRIPTION = {"model": "Model Description",
                                  "training": "Training Parameters",
                                  "training_management": "Training Management and Monitoring"}


def define_parser(parser):
    parser.add_argument("data_prefix", nargs="?",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                        help="prefix of the training data created by make_data.py")
    parser.add_argument("save_prefix", nargs="?",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                        help="prefix to be added to all files created during the training")

    model_description_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["model"])
    model_description_group.add_argument("--Ei", type=int, default=640, help="Source words embedding size.")
    model_description_group.add_argument("--Eo", type=int, default=640, help="Target words embedding size.")
    model_description_group.add_argument("--Hi", type=int, default=1024, help="Source encoding layer size.")
    model_description_group.add_argument("--Ho", type=int, default=1024, help="Target hidden layer size.")
    model_description_group.add_argument("--Ha", type=int, default=1024, help="Attention Module Hidden layer size.")
    model_description_group.add_argument("--Hl", type=int, default=512, help="Maxout output size.")
    model_description_group.add_argument("--encoder_cell_type", default="lstm", help="cell type of encoder. format: type,param1:val1,param2:val2,...")  # where type is in [%s]"%(" ".join(rnn_cells.cell_dict.keys())))
    model_description_group.add_argument("--decoder_cell_type", default="lstm", help="cell type of decoder. format same as for encoder")
    model_description_group.add_argument("--lexical_probability_dictionary", help="lexical translation probabilities in zipped JSON format. Used to implement https://arxiv.org/abs/1606.02006")
    model_description_group.add_argument("--lexicon_prob_epsilon", default=1e-3, type=float, help="epsilon value for combining the lexical probabilities")
    model_description_group.add_argument("--use_deep_attn", default=False, action="store_true")
    model_description_group.add_argument("--use_accumulated_attn", default=False, action="store_true")
    model_description_group.add_argument("--init_orth", default=False, action="store_true")
    model_description_group.add_argument("--use_bn_length", default=0, type=int)

    training_paramenters_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["training"])
    training_paramenters_group.add_argument("--mb_size", type=int, default=64, help="Minibatch size")
    training_paramenters_group.add_argument("--nb_batch_to_sort", type=int, default=20, help="Sort this many batches by size.")
    training_paramenters_group.add_argument("--noise_on_prev_word", default=False, action="store_true")
    training_paramenters_group.add_argument("--l2_gradient_clipping", type=float, default=1, help="L2 gradient clipping. 0 for None")
    training_paramenters_group.add_argument("--hard_gradient_clipping", type=float, nargs=2, help="hard gradient clipping.")
    training_paramenters_group.add_argument("--weight_decay", type=float, help="Weight decay value. ")
    training_paramenters_group.add_argument("--optimizer", choices=["sgd", "rmsprop", "rmspropgraves",
                                                                    "momentum", "nesterov", "adam", "adagrad", "adadelta"],
                                            default="adam", help="Optimizer type.")
    training_paramenters_group.add_argument("--learning_rate", type=float, default=0.01, help="Learning Rate")
    training_paramenters_group.add_argument("--momentum", type=float, default=0.9, help="Momentum term")
    training_paramenters_group.add_argument("--randomized_data", default=False, action="store_true")
    training_paramenters_group.add_argument("--no_shuffle_of_training_data", default=False, action="store_true")
    training_paramenters_group.add_argument("--use_reinf", default=False, action="store_true")
    training_paramenters_group.add_argument("--use_previous_prediction", default=0, type=float)
    training_paramenters_group.add_argument("--curiculum_training", default=False, action="store_true")
    training_paramenters_group.add_argument("--reverse_src", default=False, action="store_true")
    training_paramenters_group.add_argument("--reverse_tgt", default=False, action="store_true")

    training_monitoring_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["training_management"])
    training_monitoring_group.add_argument("--config", help="load a training config file")
    training_monitoring_group.add_argument("--data_prefix", dest="data_prefix",
                                           action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                                           help="same as positional argument --data_prefix")
    training_monitoring_group.add_argument("--save_prefix", dest="save_prefix",
                                           action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone,
                                           help="same as positional argument --save_prefix")
    training_monitoring_group.add_argument("--gpu", type=int, help="specify gpu number to use, if any")
    training_monitoring_group.add_argument("--load_model", help="load the parameters of a previously trained model")
    training_monitoring_group.add_argument("--load_optimizer_state", help="load previously saved optimizer states")
    training_monitoring_group.add_argument("--load_trainer_snapshot", help="load previously saved trainer states")
    training_monitoring_group.add_argument("--use_memory_optimization", default=False, action="store_true",
                                           help="Experimental option that could strongly reduce memory used.")
    training_monitoring_group.add_argument("--max_nb_iters", type=int, default=None, help="maximum number of iterations")
    training_monitoring_group.add_argument("--max_nb_epochs", type=int, default=None, help="maximum number of epochs")
    training_monitoring_group.add_argument("--max_src_tgt_length", type=int, help="Limit length of training sentences")
    training_monitoring_group.add_argument("--report_every", type=int, default=200, help="report every x iterations")
    training_monitoring_group.add_argument("--no_resume", default=False, action="store_true")
    training_monitoring_group.add_argument("--no_report_or_save", default=False, action="store_true")
    training_monitoring_group.add_argument("--sample_every", default=200, type=int)
    training_monitoring_group.add_argument("--save_ckpt_every", default=4000, type=int)
    training_monitoring_group.add_argument("--save_initial_model_to", help="save the initial model parameters to given file in npz format")
    training_monitoring_group.add_argument("--reshuffle_every_epoch", default=False, action="store_true", help="reshuffle training data at the end of each epoch")
    training_monitoring_group.add_argument("--resume", default=False, action="store_true", help="resume training from checkpoint config")
    training_monitoring_group.add_argument("--timer_hook", default=False, action="store_true", help="activate timer hook for profiling")
    training_monitoring_group.add_argument("--force_overwrite", default=False, action="store_true", help="Do not ask before overwiting existing files")
    training_monitoring_group.add_argument("--description", help="Optional message to be stored in the configuration file")


class CommandLineValuesException(Exception):
    pass

#
# def load_training_config_file(filename):
#     file_content = json.load(open(filename))


def get_parse_option_orderer():
    description_to_config_section = dict((v, k) for (k, v) in _CONFIG_SECTION_TO_DESCRIPTION.iteritems())
    por = argument_parsing_tools.ParseOptionRecorder(group_title_to_section=description_to_config_section,
                                                     ignore_positional_arguments=set(["save_prefix", "data_prefix"]))
    define_parser(por)
    return por


def convert_cell_string(config_training, no_error=False):
    import nmt_chainer.models.rnn_cells_config

    try:
        if "encoder_cell_type" in config_training["model"] and config_training["model"]["encoder_cell_type"] is not None:
            config_training["model"]["encoder_cell_type"] = nmt_chainer.models.rnn_cells_config.create_cell_config_from_string(
                config_training["model"]["encoder_cell_type"])

        if "decoder_cell_type" in config_training["model"] and config_training["model"]["decoder_cell_type"] is not None:
            config_training["model"]["decoder_cell_type"] = nmt_chainer.models.rnn_cells_config.create_cell_config_from_string(
                config_training["model"]["decoder_cell_type"])
    except BaseException:
        if not no_error:
            raise


def load_config_train(filename, readonly=True, no_error=False):
    config = argument_parsing_tools.OrderedNamespace.load_from(filename)
    if "metadata" not in config:  # older config file
        parse_option_orderer = get_parse_option_orderer()
        config_training = parse_option_orderer.convert_args_to_ordered_dict(config["command_line"], args_is_namespace=False)

        convert_cell_string(config_training, no_error=no_error)

        assert "data" not in config_training
        config_training["data"] = argument_parsing_tools.OrderedNamespace()
        config_training["data"]["data_fn"] = config["data"]
        config_training["data"]["Vi"] = config["Vi"]
        config_training["data"]["Vo"] = config["Vo"]
        config_training["data"]["voc"] = config["voc"]

        assert "metadata" not in config_training
        config_training["metadata"] = argument_parsing_tools.OrderedNamespace()
        config_training["metadata"]["config_version_num"] = 0.9
        config_training["metadata"]["command_line"] = None
        config_training["metadata"]["knmt_version"] = None
        config = config_training
    elif config["metadata"]["config_version_num"] != 1.0:
        raise ValueError("The config version of %s is not supported by this version of the program" % filename)

    # Compatibility with intermediate verions of config file
    if "data_prefix" in config and "data_prefix" not in config["training_management"]:
        config["training_management"]["data_prefix"] = config["data_prefix"]
        del config["data_prefix"]

    if "train_prefix" in config and "train_prefix" not in config["training_management"]:
        config["training_management"]["train_prefix"] = config["train_prefix"]
        del config["train_prefix"]

    if readonly:
        config.set_readonly()
    return config


def find_which_command_line_arguments_were_given(argument_list):
    pwndan = argument_parsing_tools.ParserWithNoneDefaultAndNoGroup()
    define_parser(pwndan)
    args_given_set = pwndan.get_args_given(argument_list)
    return args_given_set


def make_config_from_args(args, readonly=True):
    config_base = None
    if args.config is not None:
        log.info("loading training config file %s", args.config)
        config_base = load_config_train(args.config, readonly=False)

    parse_option_orderer = get_parse_option_orderer()
    config_training = parse_option_orderer.convert_args_to_ordered_dict(args)

    convert_cell_string(config_training)

    if config_base is not None:
        args_given_set = find_which_command_line_arguments_were_given(
            args.__original_argument_list)
        for argname in set(args_given_set):
            if getattr(args, argname) is None:
                args_given_set.remove(argname)

        print "args_given_set", args_given_set
        config_base.update_recursive(config_training, valid_keys=args_given_set)
        config_training = config_base
    else:
        assert "data" not in config_training
        assert "metadata" not in config_training

#     config_data_fn = config_training["data_prefix"] + ".data.config"

    if config_training["training_management"]["data_prefix"] is None or config_training["training_management"]["save_prefix"] is None:
        raise CommandLineValuesException("save_prefix and data_prefix need to be set either on the command line or in a config file")

    config_training.add_metadata_infos(version_num=1, overwrite=args.config is not None)

    if readonly:
        config_training.set_readonly()

    return config_training


# def load_config_train(filename, readonly = True):
#     config_as_ordered_dict = json.load(open(filename), object_pairs_hook=OrderedDict)
#
#     config = OrderedNamespace.load_from(filename)
#     if "metadata" not in config_as_ordered_dict: # older config file
#         parse_option_orderer = get_parse_option_orderer()
#         config_training = parse_option_orderer.convert_args_to_ordered_dict(config_as_ordered_dict["command_line"], args_is_namespace = False)
#
#         assert "data" not in config_training
#         config_training["data"] = argument_parsing_tools.OrderedNamespace()
#         config_training["data"]["data_fn"] = config_as_ordered_dict["data"]
#         config_training["data"]["Vi"] = config_as_ordered_dict["Vi"]
#         config_training["data"]["Vo"] = config_as_ordered_dict["Vo"]
#         config_training["data"]["voc"] = config_as_ordered_dict["voc"]
#
#         assert "metadata" not in config_training
#         config_training["metadata"] = argument_parsing_tools.OrderedNamespace()
#         config_training["metadata"]["config_version_num"] = 0.9
#         config_training["metadata"]["command_line"] = None
#         config_training["metadata"]["knmt_version"] = None
#     elif config_as_ordered_dict["metadata"]["config_version_num"] == 1.0:
#         argument_parsing_tools.OrderedNamespace.convert_to_ordered_namespace(config_as_ordered_dict)
#         config_training = config_as_ordered_dict
#     else:
#         raise ValueError("The config version of %s is not supported by this version of the program" % filename)
#
#     if readonly:
#         config_training.set_readonly()
#     return config_training

def command_line(arguments=None):
    import argparse
    parser = argparse.ArgumentParser(description="Train a RNNSearch model",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    define_parser(parser)
    args = parser.parse_args(args=arguments)

    do_train(args)


def do_train(args):
    import nmt_chainer.training_module.train
    config = make_config_from_args(args, readonly=False)
    nmt_chainer.training_module.train.do_train(config)


if __name__ == '__main__':
    command_line()



"""eval_config.py: parse evaluation argument and create configuration dictionnary"""

from nmt_chainer.utilities import argument_parsing_tools

_CONFIG_SECTION_TO_DESCRIPTION = {"method": "Translation Method",
                                  "output": "Output Options",
                                  "process": "Translation Process Options"}


def define_parser(parser):
    parser.add_argument("training_config", nargs="?", help="prefix of the trained model",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    parser.add_argument("trained_model", nargs="?", help="prefix of the trained model",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    parser.add_argument("src_fn", nargs="?", help="source text",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    parser.add_argument("dest_fn", nargs="?", help="destination file",
                        action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)

    translation_method_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["method"])
    translation_method_group.add_argument("--mode", default="translate",
                                          choices=["translate", "align", "translate_attn", "beam_search", "eval_bleu",
                                                   "score_nbest"], help="target text")
    translation_method_group.add_argument("--beam_width", type=int, default=20, help="beam width")
    translation_method_group.add_argument("--beam_pruning_margin", type=float, default=None, help="beam pruning margin")
    parser.add_argument("--beam_score_coverage_penalty", choices=['none', 'google'], default='none')
    parser.add_argument("--beam_score_coverage_penalty_strength", type=float, default=0.2)
    translation_method_group.add_argument("--nb_steps", type=int, default=50, help="nb_steps used in generation")
    translation_method_group.add_argument("--nb_steps_ratio", type=float, help="nb_steps used in generation as a ratio of input length")
#     translation_method_group.add_argument("--beam_opt", default = False, action = "store_true")
    translation_method_group.add_argument("--groundhog", default=False, action="store_true")
    translation_method_group.add_argument("--force_finish", default=False, action="store_true")
    translation_method_group.add_argument("--beam_score_length_normalization", choices=['none', 'simple', 'google'], default='none')
    translation_method_group.add_argument("--beam_score_length_normalization_strength", type=float, default=0.2)
    translation_method_group.add_argument("--post_score_length_normalization", choices=['none', 'simple', 'google'], default='simple')
    translation_method_group.add_argument("--post_score_length_normalization_strength", type=float, default=0.2)
    translation_method_group.add_argument("--post_score_coverage_penalty", choices=['none', 'google'], default='none')
    translation_method_group.add_argument("--post_score_coverage_penalty_strength", type=float, default=0.2)
    translation_method_group.add_argument("--prob_space_combination", default=False, action="store_true")

    output_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["output"])
    output_group.add_argument("--tgt_fn", help="target text")
    output_group.add_argument("--nbest_to_rescore", help="nbest list in moses format")
    output_group.add_argument("--ref", help="target text")
    output_group.add_argument("--tgt_unk_id", choices=["align", "id"], default="align")
    output_group.add_argument("--generate_attention_html", help="generate a html file with attention information")
    output_group.add_argument("--rich_output_filename", help="generate a JSON file with attention information")
    # arguments for unk replace
    output_group.add_argument("--dic")
    output_group.add_argument("--remove_unk", default=False, action="store_true")
    output_group.add_argument("--normalize_unicode_unk", default=False, action="store_true")
    output_group.add_argument("--attempt_to_relocate_unk_source", default=False, action="store_true")

    management_group = parser.add_argument_group(_CONFIG_SECTION_TO_DESCRIPTION["process"])
    management_group.add_argument("--gpu", type=int, help="specify gpu number to use, if any")
    management_group.add_argument("--max_nb_ex", type=int, help="only use the first MAX_NB_EX examples")
    management_group.add_argument("--mb_size", type=int, default=80, help="Minibatch size")
    management_group.add_argument("--nb_batch_to_sort", type=int, default=20, help="Sort this many batches by size.")
    management_group.add_argument("--load_model_config", nargs="+", help="gives a list of models to be used for translation")
    management_group.add_argument("--src_fn", nargs="?", help="source text",
                                  action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    management_group.add_argument("--dest_fn", nargs="?", help="destination file",
                                  action=argument_parsing_tools.ArgumentActionNotOverwriteWithNone)
    management_group.add_argument("--additional_training_config", nargs="*", help="prefix of the trained model")
    management_group.add_argument("--additional_trained_model", nargs="*", help="prefix of the trained model")
    management_group.add_argument("--reverse_training_config", help="prefix of the trained model")
    management_group.add_argument("--reverse_trained_model", help="prefix of the trained model")
#     management_group.add_argument("--config", help = "load eval config file")
    management_group.add_argument("--server", help="host:port for listening request")
    management_group.add_argument("--segmenter_command", help="command to communicate with the segmenter server")
    management_group.add_argument("--segmenter_format", help="format to expect from the segmenter (parse_server, morph)", default='plain')
    management_group.add_argument("--description", help="Optional message to be stored in the configuration file")


class CommandLineValuesException(Exception):
    pass


def get_parse_option_orderer():
    description_to_config_section = dict((v, k) for (k, v) in _CONFIG_SECTION_TO_DESCRIPTION.iteritems())
    por = argument_parsing_tools.ParseOptionRecorder(group_title_to_section=description_to_config_section,
                                                     # ignore_positional_arguments = set(["src_fn", "dest_fn"])
                                                     )
    define_parser(por)
    return por


def make_config_eval(args):
    parse_option_orderer = get_parse_option_orderer()
    config_eval = parse_option_orderer.convert_args_to_ordered_dict(args)
    config_eval.add_metadata_infos(version_num=1)
    config_eval.set_readonly()

    if config_eval.process.server is None:
        if config_eval.process.src_fn is None or config_eval.process.dest_fn is None:
            raise CommandLineValuesException("src_fn and dest_fn need to be set either on the command line or in a config file")

    return config_eval


def load_config_eval(filename, readonly=True):
    config = argument_parsing_tools.OrderedNamespace.load_from(filename)
    if "metadata" not in config:  # older config file
        parse_option_orderer = get_parse_option_orderer()
        config_eval = parse_option_orderer.convert_args_to_ordered_dict(config, args_is_namespace=False)
        assert "metadata" not in config_eval
        config_eval["metadata"] = argument_parsing_tools.OrderedNamespace()
        config_eval["metadata"]["config_version_num"] = 0.9
        config_eval["metadata"]["command_line"] = None
        config_eval["metadata"]["knmt_version"] = None
        config = config_eval
    elif config["metadata"]["config_version_num"] != 1.0:
        raise ValueError("The config version of %s is not supported by this version of the program" % filename)
    if readonly:
        config.set_readonly()
    return config


def command_line(arguments=None):

    import argparse
    parser = argparse.ArgumentParser(description="Use a RNNSearch model",
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    define_parser(parser)

    args = parser.parse_args(args=arguments)

    do_eval(args)


def do_eval(args):
    import nmt_chainer.translation.eval
    config_eval = make_config_eval(args)

    if config_eval.training_config is not None:
        if config_eval.trained_model is None:
            raise CommandLineValuesException(
                "If specifying a model via the training_config argument, you also need to specify the trained_model argument")
    else:
        if config_eval.process.load_model_config is None:
            raise CommandLineValuesException(
                "You need to specify either the training_config positional argument, or the load_model_config option, or both")

    nmt_chainer.translation.eval.do_eval(config_eval)


if __name__ == '__main__':
    command_line()



#!/usr/bin/env python
"""aligned_parse_reader.py: read parsed and aligned files"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import logging
import codecs
import exceptions

log = logging.getLogger("aparse")
log.setLevel(logging.INFO)


def read_one_parse_info_from_file_object(f):
    id_line = f.readline()
    if len(id_line) == 0:
        raise exceptions.EOFError()
    sharp, id_part, score_part = id_line.split()
    assert sharp == "#"
    id_tag, id_ = id_part.split("=")
    assert id_tag == "ID"
    score_tag, score = score_part.split("=")
    assert score_tag == "SCORE"
    score = float(score)
    sentence = []
    while True:
        line = f.readline().strip()
        if len(line) == 0:
            return id_, sentence
        splitted_line = line.split("\t")
        num_pos = int(splitted_line[0])
        dpnd = int(splitted_line[1])
        word = splitted_line[2]
        assert num_pos == len(sentence)
        sentence.append(word)


def read_one_align_info_from_file_object(f):
    id_line = f.readline()
    if len(id_line) == 0:
        raise exceptions.EOFError()
    id_line = id_line.strip()
    sharp, id_, score = id_line.split()
    assert sharp == "#"
    score = float(score)
    align_line = f.readline().strip()
    alignment = []
    for link in align_line.split():
        left, right = link.split("-")
        left = [int(x) for x in left.split(",")]
        right = [int(x) for x in right.split(",")]
        alignment.append((left, right))
    return id_, score, alignment


def load_aligned_corpus(src_fn, tgt_fn, align_fn, skip_empty_align=True, invert_alignment_links=False):
    src = codecs.open(src_fn, encoding="utf8")
    tgt = codecs.open(tgt_fn, encoding="utf8")
    align_f = codecs.open(align_fn, encoding="utf8")

    num_sentence = 0
    while True:
        try:
            id_src, sentence_src = read_one_parse_info_from_file_object(src)
            id_tgt, sentence_tgt = read_one_parse_info_from_file_object(tgt)
            id_align, score_align, alignment = read_one_align_info_from_file_object(
                align_f)
        except exceptions.EOFError:
            return
        if skip_empty_align and len(alignment) == 0:
            log.warn("skipping empty alignment %i %s" % (num_sentence, id_align))
            continue
        assert id_src == id_tgt, "%s != %s @%i" % (id_src, id_tgt, num_sentence)
        assert id_src == id_align, "%s != %s @%i" % (id_src, id_align, num_sentence)

        if invert_alignment_links:
            inverted_alignment = [(right, left) for (left, right) in alignment]
            alignment = inverted_alignment

        yield sentence_src, sentence_tgt, alignment
        num_sentence += 1



import logging
import itertools
import subprocess
import time

logging.basicConfig()
log = logging.getLogger("multirun")
log.setLevel(logging.INFO)

def parse_var(var_string):
    if var_string is None:
        return None
    groups_strings = var_string.split("|")
    value_set = [x.split("&") for x in groups_strings]
    return value_set
    
def generate_all_values(var_values):
    """
    var_values is a list of value_set, one for each var
    value_set can be set to None
    value_set is a list of groups; a group is a list of values
    """
    
    nb_groups = None
    var_set = []
    for num_var, value_set in enumerate(var_values):
        if value_set is None:
            continue
        var_set.append(num_var)
        if nb_groups is None:
            nb_groups = len(value_set)
        elif nb_groups != len(value_set):
            raise ValueError("var %i as a number of groups incompatible with previous var")
        
    set_var_values = [var_values[i] for i in var_set]
        
    result = []
    for group_list in zip(*set_var_values):
        for vars_val in itertools.product(*group_list):
            full_vars_val = [""] * len(var_values)
            for var_num, vv in zip(var_set, vars_val):
                full_vars_val[var_num] = vv
            result.append(full_vars_val)
        
    return result
    
MAX_NB_VARS = 5

def define_parser(parser):
    for num_var in range(MAX_NB_VARS):
        parser.add_argument("--var%i"%num_var)
    parser.add_argument("cmd")
    
def do_run(args):
    var_values = []
    for num_var in range(MAX_NB_VARS):
        var_values.append(parse_var(getattr(args, "var%i"%num_var)))
    
    for vals in generate_all_values(var_values):
        cmd = args.cmd.format(*vals)
        print "[%s] LAUNCHING: %s"%(time.strftime("%c"), cmd)
        subprocess.call(cmd, shell = True)
        print "[%s] FINISHED: %s"%(time.strftime("%c"), cmd)

def command_line():
    import argparse
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    define_parser(parser)
    args = parser.parse_args()
    do_run(args)
    
if __name__ == "__main__":
    command_line()



#!/usr/bin/env python
"""attention.py: Implementation of Attention mechanisms"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L

from nmt_chainer.utilities.utils import ortho_init

import logging
logging.basicConfig()
log = logging.getLogger("rnns:attn")
log.setLevel(logging.INFO)


class AttentionModule(Chain):
    """ Attention Module for computing the current context during decoding.
        The __call_ takes 2 parameters: fb_concat and mask.

        fb_concat should be the result of a call to Encoder.
        mask is as in the description of Encoder

        Return a chainer variable of shape (mb_size, Hi) and type float32
    """

    def __init__(self, Hi, Ha, Ho, init_orth=False):
        super(AttentionModule, self).__init__(
            al_lin_h=L.Linear(Hi, Ha, nobias=False),
            al_lin_s=L.Linear(Ho, Ha, nobias=True),
            al_lin_o=L.Linear(Ha, 1, nobias=True)
        )
        self.Hi = Hi
        self.Ha = Ha

        if init_orth:
            ortho_init(self.al_lin_h)
            ortho_init(self.al_lin_s)
            ortho_init(self.al_lin_o)

    def __call__(self, fb_concat, mask):
        mb_size, nb_elems, Hi = fb_concat.data.shape
        assert Hi == self.Hi
        precomputed_al_factor = F.reshape(self.al_lin_h(
            F.reshape(fb_concat, (mb_size * nb_elems, self.Hi))), (mb_size, nb_elems, self.Ha))

        mask_length = len(mask)
        seq_length = nb_elems
        assert mask_length <= seq_length
        mask_offset = seq_length - mask_length

#         concatenated_mask = F.concat([F.reshape(mask_elem, (mb_size, 1)) for mask_elem in mask], 1)

        if mask_length > 0:
            with cuda.get_device(mask[0]):
                if mask_offset > 0:
                    concatenated_penalties = self.xp.concatenate(
                        [
                            self.xp.zeros((mb_size, mask_offset), dtype=self.xp.float32),
                            -10000 * (1 - self.xp.concatenate([
                                self.xp.reshape(mask_elem, (mb_size, 1)).astype(self.xp.float32) for mask_elem in mask], 1))
                        ], 1
                    )
                else:
                    concatenated_penalties = -10000 * (1 - self.xp.concatenate([
                        self.xp.reshape(mask_elem, (mb_size, 1)).astype(self.xp.float32) for mask_elem in mask], 1))

        def compute_ctxt(previous_state):
            current_mb_size = previous_state.data.shape[0]
            if current_mb_size < mb_size:
                al_factor, _ = F.split_axis(
                    precomputed_al_factor, (current_mb_size,), 0)
                used_fb_concat, _ = F.split_axis(
                    fb_concat, (current_mb_size,), 0)
                if mask_length > 0:
                    used_concatenated_penalties = concatenated_penalties[:current_mb_size]
            else:
                al_factor = precomputed_al_factor
                used_fb_concat = fb_concat
                if mask_length > 0:
                    used_concatenated_penalties = concatenated_penalties

            state_al_factor = self.al_lin_s(previous_state)
            state_al_factor_bc = F.broadcast_to(F.reshape(state_al_factor, (current_mb_size, 1, self.Ha)), (current_mb_size, nb_elems, self.Ha))
            a_coeffs = F.reshape(self.al_lin_o(F.reshape(F.tanh(state_al_factor_bc + al_factor),
                                                         (current_mb_size * nb_elems, self.Ha))), (current_mb_size, nb_elems))

            if mask_length > 0:
                with cuda.get_device(used_concatenated_penalties):
                    a_coeffs = a_coeffs + used_concatenated_penalties  # - 10000 * (1-used_concatenated_mask.data)

            attn = F.softmax(a_coeffs)

            ci = F.reshape(F.batch_matmul(attn, used_fb_concat, transa=True), (current_mb_size, self.Hi))

            return ci, attn

        return compute_ctxt

    def compute_ctxt_demux(self, fb_concat, mask):
        mb_size, nb_elems, Hi = fb_concat.data.shape
        assert Hi == self.Hi
        assert mb_size == 1
        assert len(mask) == 0

        precomputed_al_factor = F.reshape(self.al_lin_h(
            F.reshape(fb_concat, (mb_size * nb_elems, self.Hi))), (mb_size, nb_elems, self.Ha))

#         concatenated_mask = F.concat([F.reshape(mask_elem, (mb_size, 1)) for mask_elem in mask], 1)

        def compute_ctxt(previous_state):
            current_mb_size = previous_state.data.shape[0]

            al_factor = F.broadcast_to(precomputed_al_factor, (current_mb_size, nb_elems, self.Ha))
#             used_fb_concat = F.broadcast_to(fb_concat, (current_mb_size, nb_elems, Hi))
#             used_concatenated_mask = F.broadcast_to(concatenated_mask, (current_mb_size, nb_elems))

            state_al_factor = self.al_lin_s(previous_state)
            state_al_factor_bc = F.broadcast_to(F.reshape(state_al_factor, (current_mb_size, 1, self.Ha)), (current_mb_size, nb_elems, self.Ha))
            a_coeffs = F.reshape(self.al_lin_o(F.reshape(F.tanh(state_al_factor_bc + al_factor),
                                                         (current_mb_size * nb_elems, self.Ha))), (current_mb_size, nb_elems))


#             with cuda.get_device(used_concatenated_mask.data):
#                 a_coeffs = a_coeffs - 10000 * (1-used_concatenated_mask.data)

            attn = F.softmax(a_coeffs)

#             ci = F.reshape(F.batch_matmul(attn, used_fb_concat, transa = True), (current_mb_size, self.Hi))

            ci = F.reshape(F.matmul(attn, F.reshape(fb_concat, (nb_elems, Hi))), (current_mb_size, self.Hi))

            return ci, attn

        return compute_ctxt


class DeepAttentionModule(Chain):
    """ DeepAttention Module for computing the current context during decoding.
        The __call_ takes 2 parameters: fb_concat and mask.

        fb_concat should be the result of a call to Encoder.
        mask is as in the description of Encoder

        Return a chainer variable of shape (mb_size, Hi) and type float32
    """

    def __init__(self, Hi, Ha, Ho, init_orth=False):
        log.info("using deep attention")
        super(DeepAttentionModule, self).__init__(
            attn1=AttentionModule(Hi, Ha, Ho, init_orth=init_orth),
            attn2=AttentionModule(Hi, Ha, Ho + Hi, init_orth=init_orth)
        )

    def __call__(self, fb_concat, mask):
        compute_ctxt1 = self.attn1(fb_concat, mask)
        compute_ctxt2 = self.attn2(fb_concat, mask)

        def compute_ctxt(previous_state):
            ci1, attn1 = compute_ctxt1(previous_state)
            intermediate_state = F.concat((previous_state, ci1), axis=1)
            ci2, attn2 = compute_ctxt2(intermediate_state)

            return ci2, attn2

        return compute_ctxt

    def compute_ctxt_demux(self, fb_concat, mask):
        raise NotImplemented


class CopyMechanism(Chain):
    def __init__(self, Hi, Ho):
        super(CopyMechanism, self).__init__(
            lin=L.Linear(Hi, Ho)
        )
        self.Hi = Hi
        self.Ho = Ho

    def __call__(self, inpt, mask):
        mb_size = inpt.data.shape[0]
        max_length = inpt.data.shape[1]

        precomp = F.reshape(F.tanh(self.lin(F.reshape(inpt, (-1, self.Hi)))), (mb_size, -1, self.Ho))

        mask_offset = max_length - len(mask)

        precomp_mask_penalties = self.xp.concatenate(
            [
                self.xp.zeros((mb_size, mask_offset), dtype=self.xp.float32),
                -10000 * (1 - self.xp.concatenate([
                    self.xp.reshape(mask_elem, (mb_size, 1)).astype(self.xp.float32) for mask_elem in mask], 1))
            ], 1
        )

        def compute_copy_coefficients(state):
            betas = F.reshape(F.batch_matmul(precomp, state), (mb_size, -1))
            masked_betas = betas + precomp_mask_penalties
            return masked_betas

        return compute_copy_coefficients



#!/usr/bin/env python
"""graph_training.py: visualize training process"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import sqlite3


def build_prefix_list(lst, op=max):
    res = [None] * len(lst)
    res[0] = lst[0]
    for k in xrange(1, len(lst)):
        res[k] = op(res[k - 1], lst[k])
    return res


def define_parser(parser):
    parser.add_argument("sqldb")
    parser.add_argument("dest")
#     parser.add_argument("--lib", default = "plotly", choices = ["bokeh", "plotly"])


def do_graph(args):
    generate_graph(args.sqldb, args.dest)


def commandline():
    import argparse
    parser = argparse.ArgumentParser()
    define_parser(parser)
    args = parser.parse_args()
    do_graph(args)


def generate_graph(sqldb, dest, title="Training Evolution"):
    db = sqlite3.connect(sqldb)
    c = db.cursor()

    c.execute("SELECT date, bleu_info, iteration, loss, bleu, dev_loss, dev_bleu, avg_time, avg_training_loss FROM exp_data")
    date, bleu_info, iteration, test_loss, test_bleu, dev_loss, dev_bleu, avg_time, avg_training_loss = zip(
        *list(c.fetchall()))
#     all = [[], [], [], []]
#     for line in c.fetchall():
#         print line
# #         _, _, _, test_loss, test_bleu, dev_loss, dev_bleu = line
#         test_loss, test_bleu, dev_loss, dev_bleu, avg_training_loss = line
#         all[0].append(test_loss)
#         all[1].append(test_bleu)
#         all[2].append(dev_loss)
#         all[3].append(dev_bleu)
    if avg_training_loss[0] == 0:
        avg_training_loss = list(avg_training_loss)
        avg_training_loss[0] = None

    import plotly
    #         import plotly.plotly as py
    import plotly.graph_objs as go
    import numpy as np
    #         N = len(all[0])
    # #         random_x = np.linspace(0, , N)
    #         random_y0 = np.random.randn(N)+5
    #         random_y1 = np.random.randn(N)
    #         random_y2 = np.random.randn(N)-5
    #
    # Create traces
    text_trg = ["i:%i\n%s\nt:%r" % (i, d, t) for i, d, t in zip(iteration, date, avg_time)]
    text_bleu = []
    for bli in bleu_info:
        if bli is None:
            text_bleu.append("NO INFO")
        else:
            bli_splitted = bli.split()
            ng_info = bli_splitted[1:5]
            size_info = bli_splitted[-1]
            text_bleu.append("\n" + "\n".join(ng_info + [size_info]))

    trace0 = go.Scatter(
        #             x = random_x,
        y=test_loss,
        mode='lines',
        name='test_loss',
        line=dict(
            color=('rgb(205, 12, 24)'),
            width=2)
    )

    trace0min = go.Scatter(
        #             x = random_x,
        y=build_prefix_list(test_loss, op=min),
        mode='lines',
        name='min_test_loss',
        line=dict(
            color=('rgb(205, 12, 24)'),
            width=1, dash="dot")
    )

    trace1 = go.Scatter(
        #             x = random_x,
        y=test_bleu,
        mode='markers',
        name='test_bleu',
        yaxis='y2',
        text=text_bleu,

        marker=dict(
            size=4,
            color='rgba(205, 12, 24, .8)')


        #                 line = dict(
        #         color = ('rgb(205, 12, 24)'),
        #         width = 2)
    )

    trace1max = go.Scatter(
        y=build_prefix_list(test_bleu, op=max),
        mode='lines',
        name='max_test_bleu',
        yaxis='y2',
        line=dict(
            color=('rgb(205, 12, 24)'),
            width=1,
            dash="dot")
    )

    trace2 = go.Scatter(
        y=dev_loss,
        mode='lines',
        name='dev_loss',
        line=dict(
            color=('rgb(22, 96, 167)'),
            width=2,)
    )

    trace2min = go.Scatter(
        y=build_prefix_list(dev_loss, op=min),
        mode='lines',
        name='min_dev_loss',
        line=dict(
            color=('rgb(22, 96, 167)'),
            width=1, dash="dot")
    )

    trace3 = go.Scatter(
        y=dev_bleu,
        mode='markers',
        name='dev_bleu',
        yaxis='y2',
        marker=dict(
            size=4,
            color='rgba(22, 96, 167, .8)')
        #                 line = dict(
        #         color = ('rgb(22, 96, 167)'),
        #         width = 2,)
    )

    trace3max = go.Scatter(
        y=build_prefix_list(dev_bleu, op=max),
        mode='lines',
        name='max_dev_bleu',
        yaxis='y2',
        line=dict(
            color=('rgb(22, 96, 167)'),
            width=1, dash="dot")
    )

    trace4 = go.Scatter(
        #             x = random_x,
        y=avg_training_loss,
        mode='lines',
        text=text_trg,
        name='avg_training_loss',
        line=dict(
            color=('rgb(22, 220, 67)'),
            width=2,)
    )

    layout = go.Layout(
        title=title,
        yaxis=dict(
            title='loss',
            gridcolor='#bdbdbd'
        ),
        yaxis2=dict(
            title='bleu',
            overlaying='y',
            side='right'
        )
    )

    data = [trace0, trace1, trace2, trace3, trace4, trace0min, trace1max, trace2min, trace3max]
    #         data = [trace2, trace3, trace4, trace2min, trace3max]
    fig = go.Figure(data=data, layout=layout)
    # Plot and embed in ipython notebook!
    plotly.offline.plot(fig, filename=dest, auto_open=False)


if __name__ == '__main__':
    commandline()



import subprocess
import os
from collections import OrderedDict


def get_installed_path():
    return os.path.dirname(os.path.realpath(__file__))


def get_current_git_hash():
    try:
        return subprocess.check_output(['git', 'rev-parse', 'HEAD'],
                                       cwd=get_installed_path(),
                                       stderr=subprocess.STDOUT).strip()
    except BaseException:
        return None


def get_current_git_diff():
    try:
        return subprocess.check_output(['git', 'diff'],
                                       cwd=get_installed_path(),
                                       stderr=subprocess.STDOUT)
    except BaseException:
        return None


def is_current_git_dirty():
    try:
        DEVNULL = open(os.devnull, 'wb')
        returncode = subprocess.call(['git', 'diff-index', '--quiet', 'HEAD', '--'],
                                     cwd=get_installed_path(),
                                     stdout=DEVNULL,
                                     stderr=DEVNULL)
        if returncode == 0:
            return "clean"
        elif returncode == 1:
            return "dirty"
        else:
            return "unknown"
    except BaseException:
        return "unknown"


def get_chainer_infos():
    try:
        import chainer
        result = OrderedDict([
            ("version", chainer.__version__),
            ("cuda", chainer.cuda.available),
            ("cudnn", chainer.cuda.cudnn_enabled),
        ])
        if chainer.cuda.available:
            try:
                import cupy
                cuda_version = cupy.cuda.runtime.driverGetVersion()
            except BaseException:
                cuda_version = "unavailable"
            result["cuda_version"] = cuda_version
        else:
            result["cuda_version"] = "unavailable"

        if chainer.cuda.cudnn_enabled:
            try:
                cudnn_version = chainer.cuda.cudnn.cudnn.getVersion()
            except BaseException:
                cudnn_version = "unavailable"
            result["cudnn_version"] = cudnn_version
        else:
            result["cudnn_version"] = "unavailable"

    except ImportError:
        result = OrderedDict([
            ("version", "unavailable"),
            ("cuda", "unavailable"),
            ("cudnn", "unavailable"),
            ("cuda_version", "unavailable"),
            ("cudnn_version", "unavailable")
        ])

    return result


def get_package_git_hash():
    try:
        import nmt_chainer._build
        return nmt_chainer._build.__build__
    except BaseException:
        return None


def get_package_dirty_status():
    try:
        import nmt_chainer._build
        return nmt_chainer._build.__dirty_status__
    except BaseException:
        return "unknown"


def get_package_git_diff():
    try:
        import nmt_chainer._build
#         import json
        return nmt_chainer._build.__git_diff__
    except BaseException:
        return None


def get_version_dict():
    import nmt_chainer._version
    result = OrderedDict({"package_version": nmt_chainer._version.__version__})
    current_git_hash = get_current_git_hash()
    if current_git_hash is not None:
        result["git"] = current_git_hash
        current_git_status = is_current_git_dirty()
        result["dirty_status"] = current_git_status
        if current_git_status == "dirty":
            result["diff"] = get_current_git_diff()
        result["version_from"] = "git call"
    else:
        package_git_hash = get_package_git_hash()
        if package_git_hash is not None:
            result["git"] = package_git_hash
            current_git_status = get_package_dirty_status()
            result["dirty_status"] = current_git_status
            if current_git_status == "dirty":
                result["diff"] = get_package_git_diff()
            result["version_from"] = "setup info"
        else:
            result["git"] = "unavailable"

    result["chainer"] = get_chainer_infos()
    return result


def main(options=None):
    import nmt_chainer._version
    print "package version:", nmt_chainer._version.__version__
    print "installed in:", get_installed_path()

    print "\n*********** chainer version ***********"
    chainer_infos = get_chainer_infos()
    for keyword in "version cuda cudnn cuda_version cudnn_version".split():
        print keyword, chainer_infos[keyword]

    print "\n\n********** package build info ***********"
    print "package build (git hash):", get_package_git_hash()
    package_dirty_status = get_package_dirty_status()
    if package_dirty_status == "clean":
        print "  - package git index is clean"
    elif package_dirty_status == "dirty":
        print "  - package git index is dirty"
        print "\npackage build diff (git diff):\n", get_package_git_diff()

    print "\n\n********** current version info ***********"
    print "git hash:", get_current_git_hash()

    current_dirty_status = is_current_git_dirty()

    if current_dirty_status == "clean":
        print "  - git index is clean"
    elif current_dirty_status == "dirty":
        print "  - git index is dirty"
        print "\ngit diff:\n"
        print get_current_git_diff()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""visualisation.py: visualize attention model"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import numpy as np

from bokeh.plotting import figure, output_file, show
from bokeh.models.layouts import Column
from bokeh.models import HoverTool, ColumnDataSource

# adapted from
# http://bokeh.pydata.org/en/0.7.1/tutorial/solutions/gallery/les_mis.html


def make_alignment_figure(src, tgt, alignment, title="Attention Model", toolbar_location='right', plot_width=800, plot_height=800):

    alignment = alignment[:, ::-1]
#     tgt = list(reversed(tgt))

    src = ["%i %s" % (num, x) for num, x in enumerate(src)]
    tgt = ["%s %i" % (x, num) for num, x in reversed(list(enumerate(tgt)))]

    xname = []
    yname = []
    color = []
    alpha = []

    count = []
    for i, n1 in enumerate(src):
        for j, n2 in enumerate(tgt):
            xname.append(n1)
            yname.append(n2)
            alpha.append(alignment[i, j])
            color.append('black')
            count.append(alignment[i, j])

#     for x, y, a in zip(xname, yname, alpha):
#         print x, y, a

    source = ColumnDataSource(
        data=dict(
            xname=xname,
            yname=yname,
            colors=color,
            alphas=alpha,
            count=count,
        )
    )

    # create a new figure
    p = figure(title=title,
               x_axis_location="above", tools="resize,hover", toolbar_location=toolbar_location,
               x_range=src, y_range=tgt,
               plot_width=plot_width, plot_height=plot_height)

    p.rect('xname', 'yname', 0.9, 0.9, source=source,
           color='colors', alpha='alphas', line_color=None)

    p.grid.grid_line_color = None
    p.axis.axis_line_color = None
    p.axis.major_tick_line_color = None
    p.axis.major_label_text_font_size = "10pt"
    p.axis.major_label_standoff = 0
    p.xaxis.major_label_orientation = np.pi / 3

    hover = p.select(dict(type=HoverTool))
    hover.tooltips = [
        ('tgt/src', '@yname, @xname'),
        ('attn', '@count'),
    ]

    return p


if __name__ == '__main__':
    src = ["le", "est", "un"]
    tgt = ["the", "one", "is", "about"]
    alignment = np.array(
        [[0, 1, 2, 5],
         [2, 4, 6, 7],
         [9, 1, 3, 3]]
    ) / 10.0
    print alignment
    p1 = make_alignment_figure(src, tgt, alignment)
    p2 = make_alignment_figure(src, tgt, alignment)
    p_all = Column(p1, p2)
    show(p)



#!/usr/bin/env python
"""utils.py: Various utilitity functions for RNNSearch"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import os
import logging
import numpy as np
import chainer
from chainer import Variable, cuda
import random

logging.basicConfig()
log = logging.getLogger("rnns:utils")
log.setLevel(logging.INFO)


def ensure_path(path):
    try:
        os.makedirs(path)
        log.info("Created directory %s" % path)
    except OSError:
        if not os.path.isdir(path):
            raise


def make_batch_src(src_data, padding_idx=0, gpu=None, volatile="off"):
    max_src_size = max(len(x) for x in src_data)
    min_src_size = min(len(x) for x in src_data)
    mb_size = len(src_data)

    src_batch = [np.empty((mb_size,), dtype=np.int32) for _ in xrange(max_src_size)]
    src_mask = [np.empty((mb_size,), dtype=np.bool) for _ in xrange(max_src_size - min_src_size)]

    for num_ex in xrange(mb_size):
        this_src_len = len(src_data[num_ex])
        for i in xrange(max_src_size):
            if i < this_src_len:
                src_batch[i][num_ex] = src_data[num_ex][i]
                if i >= min_src_size:
                    src_mask[i - min_src_size][num_ex] = True
            else:
                src_batch[i][num_ex] = padding_idx
                assert i >= min_src_size
                src_mask[i - min_src_size][num_ex] = False

    if gpu is not None:
        return ([Variable(cuda.to_gpu(x, gpu), volatile=volatile) for x in src_batch],
                [cuda.to_gpu(x, gpu) for x in src_mask])
    else:
        return [Variable(x, volatile=volatile) for x in src_batch], src_mask


def make_batch_src_tgt(training_data, eos_idx=1, padding_idx=0, gpu=None, volatile="off", need_arg_sort=False):
    if need_arg_sort:
        training_data_with_argsort = zip(training_data, range(len(training_data)))
        training_data_with_argsort.sort(key=lambda x: len(x[0][1]), reverse=True)
        training_data, argsort = zip(*training_data_with_argsort)
    else:
        training_data = sorted(training_data, key=lambda x: len(x[1]), reverse=True)
#     max_src_size = max(len(x) for x, y  in training_data)
    max_tgt_size = max(len(y) for _, y in training_data)
    mb_size = len(training_data)

    src_batch, src_mask = make_batch_src(
        [x for x, y in training_data], padding_idx=padding_idx, gpu=gpu, volatile=volatile)

    lengths_list = []
    lowest_non_finished = mb_size - 1
    for pos in xrange(max_tgt_size + 1):
        while pos > len(training_data[lowest_non_finished][1]):
            lowest_non_finished -= 1
            assert lowest_non_finished >= 0
        mb_length_at_this_pos = lowest_non_finished + 1
        assert len(
            lengths_list) == 0 or mb_length_at_this_pos <= lengths_list[-1]
        lengths_list.append(mb_length_at_this_pos)

    tgt_batch = []
    for i in xrange(max_tgt_size + 1):
        current_mb_size = lengths_list[i]
        assert current_mb_size > 0
        tgt_batch.append(np.empty((current_mb_size,), dtype=np.int32))
        for num_ex in xrange(current_mb_size):
            assert len(training_data[num_ex][1]) >= i
            if len(training_data[num_ex][1]) == i:
                tgt_batch[-1][num_ex] = eos_idx
            else:
                tgt_batch[-1][num_ex] = training_data[num_ex][1][i]

    if gpu is not None:
        tgt_batch_v = [Variable(cuda.to_gpu(x, gpu), volatile=volatile) for x in tgt_batch]
    else:
        tgt_batch_v = [Variable(x, volatile=volatile) for x in tgt_batch]

    if need_arg_sort:
        return src_batch, tgt_batch_v, src_mask, argsort
    else:
        return src_batch, tgt_batch_v, src_mask


def make_batch_tgt(training_data, eos_idx=1, gpu=None, volatile="off", need_arg_sort=False):
    if need_arg_sort:
        training_data_with_argsort = zip(training_data, range(len(training_data)))
        training_data_with_argsort.sort(key=lambda x: len(x[0]), reverse=True)
        training_data, argsort = zip(*training_data_with_argsort)
    else:
        training_data = sorted(training_data, key=lambda x: len(x), reverse=True)
#     max_src_size = max(len(x) for x, y  in training_data)
    max_tgt_size = max(len(y) for y in training_data)
    mb_size = len(training_data)

    lengths_list = []
    lowest_non_finished = mb_size - 1
    for pos in xrange(max_tgt_size + 1):
        while pos > len(training_data[lowest_non_finished]):
            lowest_non_finished -= 1
            assert lowest_non_finished >= 0
        mb_length_at_this_pos = lowest_non_finished + 1
        assert len(
            lengths_list) == 0 or mb_length_at_this_pos <= lengths_list[-1]
        lengths_list.append(mb_length_at_this_pos)

    tgt_batch = []
    for i in xrange(max_tgt_size + 1):
        current_mb_size = lengths_list[i]
        assert current_mb_size > 0
        tgt_batch.append(np.empty((current_mb_size,), dtype=np.int32))
        for num_ex in xrange(current_mb_size):
            assert len(training_data[num_ex]) >= i
            if len(training_data[num_ex]) == i:
                tgt_batch[-1][num_ex] = eos_idx
            else:
                tgt_batch[-1][num_ex] = training_data[num_ex][i]

    if gpu is not None:
        tgt_batch_v = [Variable(cuda.to_gpu(x, gpu), volatile=volatile) for x in tgt_batch]
    else:
        tgt_batch_v = [Variable(x, volatile=volatile) for x in tgt_batch]

    if need_arg_sort:
        return tgt_batch_v, argsort
    else:
        return tgt_batch_v


def minibatch_looper_random(data, mb_size):
    while True:
        training_data_sampled = [None] * mb_size
        r = np.random.randint(0, len(data), size=(mb_size,))
        for i in range(mb_size):
            training_data_sampled[i] = data[r[i]]
        yield training_data_sampled


def minibatch_looper(data, mb_size, loop=True, avoid_copy=False):
    current_start = 0
    data_exhausted = False
    while not data_exhausted:
        if avoid_copy and len(data) >= current_start + mb_size:
            training_data_sampled = data[current_start: current_start + mb_size]
            current_start += mb_size
            if current_start >= len(data):
                if loop:
                    current_start = 0
                else:
                    data_exhausted = True
                    break
        else:
            training_data_sampled = []
            while len(training_data_sampled) < mb_size:
                remaining = mb_size - len(training_data_sampled)
                training_data_sampled += data[current_start:current_start + remaining]
                current_start += remaining
                if current_start >= len(data):
                    if loop:
                        current_start = 0
                    else:
                        data_exhausted = True
                        break

        yield training_data_sampled


def batch_sort_and_split(batch, size_parts, sort_key=lambda x: len(x[1]), inplace=False):
    if not inplace:
        batch = list(batch)
    batch.sort(key=sort_key)
    nb_mb_for_sorting = len(batch) / size_parts + (1 if len(batch) % size_parts != 0 else 0)
    for num_batch in xrange(nb_mb_for_sorting):
        mb_raw = batch[num_batch * size_parts: (num_batch + 1) * size_parts]
        yield mb_raw


def mb_reverser(mb_raw, reverse_src=False, reverse_tgt=False):
    if reverse_src or reverse_tgt:
        mb_raw_new = []
        for src_side, tgt_side in mb_raw:
            if reverse_src:
                src_side = src_side[::-1]
            if reverse_tgt:
                tgt_side = tgt_side[::-1]
            mb_raw_new.append((src_side, tgt_side))
        return mb_raw_new
    else:
        return mb_raw


def minibatch_provider_curiculum(data, eos_idx, mb_size, nb_mb_for_sorting=1, inplace_sorting=False, gpu=None,
                                 randomized=False, volatile="off", sort_key=lambda x: len(x[1]),
                                 reverse_src=False, reverse_tgt=False, starting_size=200
                                 ):
    current_size = starting_size
    while True:
        used_data = list(data[:current_size])
        random.shuffle(used_data)
        sub_mb_provider = minibatch_provider(used_data, eos_idx, mb_size, nb_mb_for_sorting, gpu=gpu, loop=False,
                                             randomized=randomized, sort_key=sort_key, volatile=volatile,
                                             inplace_sorting=inplace_sorting,
                                             reverse_src=reverse_src, reverse_tgt=reverse_tgt)

        for x in sub_mb_provider:
            yield x

        if current_size < len(data):
            current_size *= 2


def minibatch_provider(data, eos_idx, mb_size, nb_mb_for_sorting=1, loop=True, inplace_sorting=False, gpu=None,
                       randomized=False, volatile="off", sort_key=lambda x: len(x[1]),
                       reverse_src=False, reverse_tgt=False, give_raw_batch=False
                       ):
    if nb_mb_for_sorting == -1:
        assert not randomized
        assert not loop
        for mb_raw in batch_sort_and_split(data, mb_size, inplace=inplace_sorting, sort_key=sort_key):
            mb_raw = mb_reverser(mb_raw, reverse_src=reverse_src, reverse_tgt=reverse_tgt)
            src_batch, tgt_batch, src_mask = make_batch_src_tgt(mb_raw, eos_idx=eos_idx, gpu=gpu, volatile=volatile)

            if give_raw_batch:
                yield src_batch, tgt_batch, src_mask, mb_raw
            else:
                yield src_batch, tgt_batch, src_mask
    else:
        assert nb_mb_for_sorting > 0
        required_data = nb_mb_for_sorting * mb_size
        if randomized:
            #             assert not loop
            looper = minibatch_looper_random(data, required_data)
        else:
            looper = minibatch_looper(data, required_data, loop=loop, avoid_copy=False)
        for large_batch in looper:
            # ok to sort in place since minibatch_looper will return copies
            for mb_raw in batch_sort_and_split(large_batch, mb_size, inplace=True, sort_key=sort_key):
                mb_raw = mb_reverser(mb_raw, reverse_src=reverse_src, reverse_tgt=reverse_tgt)
                src_batch, tgt_batch, src_mask = make_batch_src_tgt(mb_raw, eos_idx=eos_idx, gpu=gpu, volatile=volatile)
                if give_raw_batch:
                    yield src_batch, tgt_batch, src_mask, mb_raw
                else:
                    yield src_batch, tgt_batch, src_mask


def compute_bleu_with_unk_as_wrong(references, candidates, is_unk_id, new_unk_id_ref, new_unk_id_cand):
    import bleu_computer
    assert new_unk_id_ref != new_unk_id_cand
    bc = bleu_computer.BleuComputer()
    for ref, cand in zip(references, candidates):
        ref_mod = tuple((x if not is_unk_id(x) else new_unk_id_ref)
                        for x in ref)
        cand_mod = tuple((int(x) if not is_unk_id(int(x))
                          else new_unk_id_cand) for x in cand)
        bc.update(ref_mod, cand_mod)
    return bc


def de_batch(batch, mask=None, eos_idx=None, is_variable=False, raw=False):
    """ Utility function for "de-batching".
        batch is a list of Variable/numpy/cupy of shape[0] <= mb_size

        returns a list of the sequences in the batch
    """
    res = []
    mb_size = len(batch[0].data) if is_variable else len(batch[0])
    if mask is not None:
        mask_offset = len(batch) - len(mask)
        assert mask_offset >= 0
    for sent_num in xrange(mb_size):
        assert sent_num == len(res)
        res.append([])
        for src_pos in range(len(batch)):
            current_batch_size = batch[src_pos].data.shape[0] if is_variable else batch[src_pos].shape[0]
            if (mask is None or
                    (src_pos < mask_offset or mask[src_pos - mask_offset][sent_num])) and current_batch_size > sent_num:
                #                 print sent_num, src_pos, batch[src_pos].data
                idx = batch[src_pos].data[sent_num] if is_variable else batch[src_pos][sent_num]
                if not raw:
                    idx = int(idx)
                res[sent_num].append(None)
                res[sent_num][src_pos] = idx
                if eos_idx is not None and idx == eos_idx:
                    break
    return res


def gen_ortho(shape):
    # adapted from code in the lasagne nn library
    a = np.random.randn(*shape)
    u, _, v = np.linalg.svd(a, full_matrices=False)
    # pick the one with the correct shape
    q = u if u.shape == a.shape else v
    return q.astype(np.float32)


def ortho_init(link):
    if isinstance(link, chainer.links.Linear):
        print "init ortho", link
        link.W.data[...] = gen_ortho(link.W.data.shape)
    elif isinstance(link, chainer.links.GRU):
        print "init ortho", link
        for name_lin in "W_r U_r W_z U_z W U".split(" "):
            print "case", name_lin, getattr(link, name_lin)
            ortho_init(getattr(link, name_lin))
    elif isinstance(link, chainer.links.Maxout):
        print "init ortho", link
        ortho_init(link.linear)
    else:
        raise NotImplemented


def compute_lexicon_matrix(src_batch, lexical_probability_dictionary, V_tgt):
    real_mb_size = src_batch[0].data.shape[0]
    max_source_size = len(src_batch)
    lexicon_matrix = np.zeros((real_mb_size, max_source_size, V_tgt), dtype=np.float32)
    for src_pos in xrange(max_source_size):
        # TODO: check if this is too slow
        src_batch_cpu = cuda.to_cpu(src_batch[src_pos].data)
        for num_mb in xrange(real_mb_size):
            src_idx = int(src_batch_cpu[num_mb])
            if src_idx in lexical_probability_dictionary:
                for tgt_idx, lex_prob in lexical_probability_dictionary[src_idx].iteritems(
                ):
                    lexicon_matrix[num_mb][src_pos][tgt_idx] = lex_prob
    return lexicon_matrix


def minibatch_sampling(probs):
    mb_size, nb_elems = probs.shape
    cum = probs.cumsum(axis=1)
    cum = np.minimum(cum, 1)  # due to rounding errors
    cum[:, -1] = 1  # due to rounding errors
    offsets = np.arange(mb_size)
    cum += offsets.reshape(mb_size, 1)
    r = np.random.rand(mb_size) + offsets
    samples = np.digitize(r, cum.reshape(-1))
    samples -= nb_elems * offsets
    return samples

#
# def create_encdec_from_config(config):
#     import models, rnn_cells
#     from make_data import Indexer
#     import json
#
#     voc_fn = config["voc"]
#     log.info("loading voc from %s"% voc_fn)
#     src_voc, tgt_voc = json.load(open(voc_fn))
#
#     src_indexer = Indexer.make_from_serializable(src_voc)
#     tgt_indexer = Indexer.make_from_serializable(tgt_voc)
#     tgt_voc = None
#     src_voc = None
#
#
# #     Vi = len(src_voc) + 1 # + UNK
# #     Vo = len(tgt_voc) + 1 # + UNK
#
#     Vi = len(src_indexer) # + UNK
#     Vo = len(tgt_indexer) # + UNK
#
#     Ei = config["Ei"]
#     Hi = config["Hi"]
#     Eo = config["Eo"]
#     Ho = config["Ho"]
#     Ha = config["Ha"]
#     Hl = config["Hl"]
#
#     encoder_cell_type = config.get("encoder_cell_type", "gru")
#     decoder_cell_type = config.get("decoder_cell_type", "gru")
#
#     use_bn_length = config.get("use_bn_length", None)
#
#     eos_idx = Vo
#
#     encdec = models.EncoderDecoder(Vi, Ei, Hi, Vo + 1, Eo, Ho, Ha, Hl, use_bn_length = use_bn_length,
#                                    encoder_cell_type = rnn_cells.cell_dict[encoder_cell_type],
#                                        decoder_cell_type = rnn_cells.cell_dict[decoder_cell_type])
#
#     return encdec, eos_idx, src_indexer, tgt_indexer



#!/usr/bin/env python
"""beam_search.py: Implementation of beam_search"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import numpy as np
import chainer
from chainer import cuda, Variable
from chainer import Link, Chain, ChainList
import chainer.functions as F
import chainer.links as L


import logging
logging.basicConfig()
log = logging.getLogger("rnns:beam_search")
log.setLevel(logging.INFO)


def iterate_best_score(new_scores, beam_width):
    """
    Create generator over the beam_width best scores.

    Args:
        new_scores: a numpy/cupy 2-dimensional array
        beam_width: a positive integer

    Returns:
        a generator that yields tuples (num_case, idx_in_case, score) where
            score is one of the top beam_width values of new_scores
            num_case is the row of this score in new_scores
            idx_in_case is the column of this score
    """
    nb_cases, v_size = new_scores.shape
    new_costs_flattened = cuda.to_cpu(- new_scores).ravel()

    # TODO replace wit a cupy argpartition when/if implemented
    best_idx = np.argpartition(new_costs_flattened, beam_width)[:beam_width]

    all_num_cases = best_idx / v_size
    all_idx_in_cases = best_idx % v_size

    for num in xrange(len(best_idx)):
        idx = best_idx[num]
        num_case = all_num_cases[num]
        idx_in_case = all_idx_in_cases[num]
        yield num_case, idx_in_case, new_costs_flattened[idx]


def iterate_eos_scores(new_scores, eos_idx):
    nb_cases, v_size = new_scores.shape

    for num_case in xrange(nb_cases):
        idx_in_case = eos_idx
        yield num_case, idx_in_case, cuda.to_cpu(new_scores[num_case, eos_idx])


def update_next_lists(num_case, idx_in_case, new_cost, eos_idx, new_state_ensemble, finished_translations, current_translations,
                      current_attentions,
                      next_states_list, next_words_list, next_score_list, next_normalized_score_list, next_translations_list,
                      attn_ensemble, next_attentions_list, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, need_attention=False):
    """
    Updates the lists containing the infos on translations in current beam

    Args:
        num_case: the index of the translation in current_translations we are going to try and extend
        idx_in_case: vocabulary index of the word we want to add to current_translations[num_case]
        eos_idx: value of the EOS index (so that we can chexk for equality with idx_in_case)

        new_state_ensemble: list of states as returned by compute_next_states_and_scores

        finished_translations: list of finished translations (for which EOS was generated)
                    each item in the list is a tuple (translation, score) or (translation, score, attention) if need_attention = True

        current_translations: list of unfinished translations in the current beam
                                each item is a sequence of target word index

        current_attentions: list of attentions for the current beam
                                each item corresponds to an item in current_translations

        next_states_list: the states extracted from new_state_ensemble corresponding to the translations in next_translations_list
        next_words_list: list of target word index corresponding to the last words of each translation in next_translations_list
        next_score_list: list of scores corresponding to the translations in next_translations_list
        next_translations_list: partially constructed list of translations for the next beam
            same structure as current_translations
        next_attentions_list: attention list corresponding to the translations in next_translations_list

        need_attention: if True, keep the attention

    Return:
        Returns None.
        But the lists finished_translations, next_states_list, next_words_list, next_score_list, next_translations_list, next_attentions_list
            will be updated.
    """
    if idx_in_case == eos_idx:
        if need_attention:
            finished_translations.append((current_translations[num_case],
                                          -new_cost,
                                          current_attentions[num_case]
                                          ))
        else:
            finished_translations.append((current_translations[num_case],
                                          -new_cost))
    else:
        next_states_list.append(
            [tuple([Variable(substates.data[num_case].reshape(1, -1), volatile="auto") for substates in new_state])
             for new_state in new_state_ensemble]
        )

        next_words_list.append(idx_in_case)
        next_score_list.append(-new_cost)

        # Compute the normalized score if needed.
        if beam_score_coverage_penalty == "google":
            coverage_penalty = 0
            if len(current_attentions[num_case]) > 0:
                xp = cuda.get_array_module(attn_ensemble[0].data)
                log_of_min_of_sum_over_j = xp.log(xp.minimum(
                    sum(current_attentions[num_case]), xp.array(1.0)))
                coverage_penalty = beam_score_coverage_penalty_strength * \
                    xp.sum(log_of_min_of_sum_over_j)
            normalized_score = -new_cost + coverage_penalty
            next_normalized_score_list.append(normalized_score)

        next_translations_list.append(
            current_translations[num_case] + [idx_in_case])
        if need_attention:
            xp = cuda.get_array_module(attn_ensemble[0].data)
            attn_summed = xp.zeros((attn_ensemble[0].data[0].shape), dtype=xp.float32)
            for attn in attn_ensemble:
                attn_summed += attn.data[num_case]
            attn_summed /= len(attn_ensemble)
            next_attentions_list.append(current_attentions[num_case] + [attn_summed])


def compute_next_lists(new_state_ensemble, new_scores, beam_width, beam_pruning_margin,
                       beam_score_length_normalization, beam_score_length_normalization_strength,
                       beam_score_coverage_penalty, beam_score_coverage_penalty_strength,
                       eos_idx,
                       current_translations,
                       finished_translations,
                       current_attentions,
                       attn_ensemble,
                       force_finish=False,
                       need_attention=False):
    """
        Compute the informations for the next beam.

        Args:
            new_state_ensemble: list of states as returned by compute_next_states_and_scores
            new_scores: numpy/cupy array of float32 representing the scores for each augmented translation
                             new_scores[num_case][idx] is the partial score of the partial
                             translation  current_translations[num_case] augmented with the word whose
                             target vocabulary index is idx.
            beam_width: maximum number of translations in a beam
            beam_pruning_margin: maximum difference of scores for translations in the same beam
            eos_idx: index of EOS in the target vocabulary
            current_translations: list of partial translations in the current beam
                                    each item is a sequence of target vocabulary index
            finished_translations: list of finished translations found so far (ie for which EOS was generated)
                    each item in the list is a tuple (translation, score) or (translation, score, attention) if need_attention = True
            current_attentions: attention for each of the partial translation in current_translations
            attn_ensemble: attention generated when computing the states in new_state_ensemble,
                            as generated by compute_next_states_and_scores
            force_finish: force the generation of EOS if we did not find a translation after nb_steps steps
            need_attention: if True, attention is kept

        Return:
            A tuple (next_states_list, next_words_list, next_score_list, next_translations_list, next_attentions_list)
                containing the informations for the next beam
    """
    # lists that contain infos on the current beam
    next_states_list = []  # each item is a list of list
    next_words_list = []
    next_score_list = []
    next_normalized_score_list = []
    next_translations_list = []
    next_attentions_list = []

    if force_finish:
        score_iterator = iterate_eos_scores(new_scores, eos_idx)
    else:
        score_iterator = iterate_best_score(new_scores, beam_width)

    for num_case, idx_in_case, new_cost in score_iterator:
        if len(current_translations[num_case]) > 0:
            if beam_score_length_normalization == 'simple':
                new_cost /= len(current_translations[num_case])
            elif beam_score_length_normalization == 'google':
                new_cost /= (pow((len(current_translations[num_case]) + 5), beam_score_length_normalization_strength) / pow(6, beam_score_length_normalization_strength))
        update_next_lists(num_case, idx_in_case, new_cost, eos_idx, new_state_ensemble,
                          finished_translations, current_translations, current_attentions,
                          next_states_list, next_words_list, next_score_list, next_normalized_score_list, next_translations_list,
                          attn_ensemble, next_attentions_list, beam_score_coverage_penalty, beam_score_coverage_penalty_strength, need_attention=need_attention)
        assert len(next_states_list) <= beam_width
#             if len(next_states_list) >= beam_width:
#                 break

    # Prune items that have a score worse than beam_pruning_margin below the
    # best score.
    if (beam_pruning_margin is not None and next_score_list):
        best_next_score = max(next_score_list)
        bad_score_indices = [
            idx for idx,
            elem in enumerate(next_score_list) if (
                best_next_score -
                elem > beam_pruning_margin)]

        for i in bad_score_indices[::-1]:
            del next_states_list[i]
            del next_words_list[i]
            del next_score_list[i]
            del next_translations_list[i]
            del next_attentions_list[i]
            if beam_score_coverage_penalty == "google":
                del next_normalized_score_list[i]

    # Prune items that have a normalized score worse than beam_pruning_margin
    # below the best normalized score.
    if (beam_score_coverage_penalty ==
            "google" and beam_pruning_margin is not None and next_normalized_score_list):
        best_next_normalized_score = max(next_normalized_score_list)
        bad_score_indices = [
            idx for idx,
            elem in enumerate(next_normalized_score_list) if (
                best_next_normalized_score -
                elem > beam_pruning_margin)]

        for i in bad_score_indices[::-1]:
            del next_states_list[i]
            del next_words_list[i]
            del next_score_list[i]
            del next_normalized_score_list[i]
            del next_translations_list[i]
            del next_attentions_list[i]

    return next_states_list, next_words_list, next_score_list, next_translations_list, next_attentions_list


def compute_next_states_and_scores(dec_cell_ensemble, current_states_ensemble, current_words,
                                   prob_space_combination=False):
    """
        Compute the next states and scores when giving current_words as input to the decoding cells in dec_cell_ensemble.

        Args:
            dec_cell_ensemble: list of decoder cells conditionalized on the input sentence
                if the length of this list is larger than one, then we will proceed to do ensemble decoding
            current_states_ensemble: list of decoder states
                                        one item for each decoder cell in dec_cell_ensemble
                                        each item can actually be a minibatch of states
            current_words: array of int32 representing the next words to give as input
                            if None, the decoder cell should use its BOS embedding as input
            prob_space_combination: if true, ensemble scores are combined by geometric average instead of arithmetic average

        Return:
            A tuple (combined_scores, new_state_ensemble, attn_ensemble) where:
                combined_scores is the log_probability returned by the decoder cell
                    or a combination of the log probabilities if there was more than one cell in dec_cell_ensemble
                new_state_ensemble is a list of states of same length as dec_cell_ensemble corresponding to the
                    new decoder states after the input of current_words
                attn_ensemble is a list attention generated by each decoder cell after being given the input current_words
    """
#     xp = cuda.get_array_module(dec_ensemble[0].initial_state.data)
    xp = dec_cell_ensemble[0].xp

    if current_words is not None:
        states_logits_attn_ensemble = [dec_cell(states, current_words) for (dec_cell, states) in zip(
            dec_cell_ensemble, current_states_ensemble)]
    else:
        assert all(x is None for x in current_states_ensemble)
        states_logits_attn_ensemble = [dec_cell.get_initial_logits(1) for dec_cell in dec_cell_ensemble]

    new_state_ensemble, logits_ensemble, attn_ensemble = zip(*states_logits_attn_ensemble)

    # Combine the scores of the ensembled models
    combined_scores = xp.zeros((logits_ensemble[0].data.shape), dtype=xp.float32)

    if not prob_space_combination:
        for logits in logits_ensemble:
            combined_scores += xp.log(F.softmax(logits).data)
        combined_scores /= len(dec_cell_ensemble)
    else:
        for logits in logits_ensemble:
            combined_scores += F.softmax(logits).data
        combined_scores /= len(dec_cell_ensemble)
        combined_scores = xp.log(combined_scores)

    return combined_scores, new_state_ensemble, attn_ensemble


def advance_one_step(dec_cell_ensemble, eos_idx, current_translations_states, beam_width, beam_pruning_margin,
                     beam_score_length_normalization,
                     beam_score_length_normalization_strength,
                     beam_score_coverage_penalty,
                     beam_score_coverage_penalty_strength,
                     finished_translations,
                     force_finish=False, need_attention=False,
                     prob_space_combination=False):
    """
        Generate the partial translations / decoder states in the next beam

        Args:
            dec_cell_ensemble: list of decoder cells conditionalized on the input sentence
                if the length of this list is larger than one, then we will proceed to do ensemble decoding
            eos_idx: the index of EOS element in the target vocabulary
            current_translations_states: a tuple representing the state of the current beam
                the tuple has the shape (translations, score, previous_states, previous_words, attention) where:
                    translations is a list of unfinished translations
                        each item is a sequence of target vocabulary index
                    score is a numpy/cupy array of same length as translations, giving the score for each item in translations
                    previous_states is a list of "states", with one state for each cell in dec_cell_ensemble
                        "states" $i$ in previous_states represents the state of cell $i$ in dec_cell_ensemble after
                            generating the translations in 'translations'. Each states thus actually represents up to
                            beam_width state, onr for each translation in 'translations'.
                            A value of None for a state indicate that the initial state of the decoder should be used.
                    previous_words is a numpy/cupy array of int32 containing the index of the last word of each translation
                        in 'translations'. If its value is None, it means the decoder should use its BOS embedding as input.
                    attention is the list of attentions generated for each of the translations in 'translations'
            finished_translations: list of finished translations
                each item in the list is a tuple (translation, score) or (translation, score, attention) if need_attention = True
            beam_width, beam_pruning_margin, force_finish, need_attention, prob_space_combination:
                see ensemble_beam_search documentation

        Returns:
            A tuple (translations, score, states, words, attention) similar to the input
                argument current_translations_states, but corresponding to the next beam.
    """

#     xp = cuda.get_array_module(dec_ensemble[0].initial_state.data)
    xp = dec_cell_ensemble[0].xp
    current_translations, current_scores, current_states_ensemble, current_words, current_attentions = current_translations_states

    # Compute the next states and associated next word scores
    combined_scores, new_state_ensemble, attn_ensemble = compute_next_states_and_scores(
        dec_cell_ensemble, current_states_ensemble, current_words,
        prob_space_combination=prob_space_combination)

    nb_cases, v_size = combined_scores.shape
    assert nb_cases <= beam_width

    # Add the new scores to the previous ones for each states in the beam
    new_scores = current_scores[:, xp.newaxis] + combined_scores

    # Compute the list of new translation states after pruning
    next_states_list, next_words_list, next_score_list, next_translations_list, next_attentions_list = compute_next_lists(
        new_state_ensemble, new_scores, beam_width, beam_pruning_margin,
        beam_score_length_normalization, beam_score_length_normalization_strength,
        beam_score_coverage_penalty, beam_score_coverage_penalty_strength,
        eos_idx,
        current_translations, finished_translations,
        current_attentions, attn_ensemble, force_finish=force_finish, need_attention=need_attention)

    if len(next_states_list) == 0:
        return None  # We only found finished translations

    # Create the new translation states

    next_words_array = np.array(next_words_list, dtype=np.int32)
    if xp is not np:
        next_words_array = cuda.to_gpu(next_words_array)

    concatenated_next_states_list = []
    for next_states_list_one_model in zip(*next_states_list):
        concatenated_next_states_list.append(
            tuple([F.concat(substates, axis=0) for substates in zip(*next_states_list_one_model)])
        )

    next_translations_states = (next_translations_list,
                                xp.array(next_score_list),
                                concatenated_next_states_list,
                                Variable(next_words_array, volatile="auto"),
                                next_attentions_list
                                )

    return next_translations_states


def ensemble_beam_search(model_ensemble, src_batch, src_mask, nb_steps, eos_idx,
                         beam_width=20, beam_pruning_margin=None,
                         beam_score_length_normalization=None,
                         beam_score_length_normalization_strength=0.2,
                         beam_score_coverage_penalty=None,
                         beam_score_coverage_penalty_strength=0.2,
                         need_attention=False,
                         force_finish=False,
                         prob_space_combination=False, use_unfinished_translation_if_none_found=False):
    """
    Compute translations using a beam-search algorithm.

    Args:
        model_ensemble: list of (trained) EncoderDecoder models
                    if the list contains more than one model, we do ensemble decoding : the scores given by each
                    models are combined to produce the best translation
        src_batch: input sentence in batch form, as generated by make_batch_src.
                    src_batch should only contains one sentence (ie minibatch size is one)
        src_mask: mask value returned by make_batch_src
        nb_steps: maximum length of the generated translation
        eos_idx: index of the EOS element in the target vocabulary
        beam_width: number of partial translation kept in each beam
        beam_pruning_margin: maximum score difference accepted within a beam
        need_attention: if True, will return the attention values for each translation
        force_finish: force the generation of EOS if we did not find a translation after nb_steps steps
        prob_space_combination: if true, ensemble scores are combined by geometric average instead of arithmetic average
        use_unfinished_translation_if_none_found: will ureturn unfinished translation if we did not find a translation after nb_steps steps

    Return:
        list of translations
            each item in the list is a tuple (translation, score) or (translation, score, attention) if need_attention = True
    """
    mb_size = src_batch[0].data.shape[0]
    assert len(model_ensemble) >= 1
    xp = model_ensemble[0].xp

    dec_cell_ensemble = [model.give_conditionalized_cell(src_batch, src_mask, noise_on_prev_word=False,
                                                         mode="test", demux=True) for model in model_ensemble]

    assert mb_size == 1
    # TODO: if mb_size == 1 then src_mask value unnecessary -> remove?

    finished_translations = []

    # Create the initial Translation state
    previous_states_ensemble = [None] * len(model_ensemble)

    # Current_translations_states will hold the information for the current beam
    current_translations_states = (
        [[]],  # translations
        xp.array([0]),  # scores
        previous_states_ensemble,  # previous states
        None,  # previous words
        [[]]  # attention
    )

    # Proceed with the search
    for num_step in xrange(nb_steps):
        current_translations_states = advance_one_step(
            dec_cell_ensemble,
            eos_idx,
            current_translations_states,
            beam_width,
            beam_pruning_margin,
            beam_score_length_normalization,
            beam_score_length_normalization_strength,
            beam_score_coverage_penalty,
            beam_score_coverage_penalty_strength,
            finished_translations,
            force_finish=force_finish and num_step == (nb_steps - 1),
            need_attention=need_attention,
            prob_space_combination=prob_space_combination)

        if current_translations_states is None:
            break

#     print finished_translations, need_attention

    # Return finished translations
    if len(finished_translations) == 0:
        if use_unfinished_translation_if_none_found:
            assert current_translations_states is not None
            if need_attention:
                translations, scores, _, _, attentions = current_translations_states
                finished_translations.append(
                    (translations[0], scores[0], attentions[0]))
            else:
                finished_translations.append((translations[0], scores[0]))
        else:
            if need_attention:
                finished_translations.append(([], 0, []))
            else:
                finished_translations.append(([], 0))
    return finished_translations
